import logging
import pyspedas
from pyspedas.projects.mms import fgm, fpi, feeps, hpca, edp, eis
from pyspedas.projects.themis import gmag
from pyspedas import options, tplot_options, store_data, tplot, tnames, time_double, time_string


def mms_overview_plot(
    trange=[],
    date="2022-03-23",
    duration="1",
    probe="1",
    skip_ae_idx=False,
    save_png='',
    save_eps='',
    save_svg='',
    save_pdf='',
    save_jpeg='',
    display=True,
    xsize=10,
    ysize=14,
):
    r"""
    Generate an overview plot for MMS data.

    Parameters:
    -----------
    trange (list of 2 strings, optional):
        Time range, start and end date. If not specified, it will be calculated based on the date and duration parameters.
    date: str, optional
        Date in the format 'YYYY-MM-DD' or 'YYY-MM-DD/hh:mm:ss. Used to calculate the time range if trange is not specified.
    duration: str, optional
        Duration in days. Used to calculate the time range if trange is not specified.
    probe: str, optional
        MMS probe number (1-4).
    skip_ae_idx: bool, optional
        Whether to skip plotting the AE index panel.
    save_png: str, optional
        .Filename to save plot as PNG
    save_eps: str, optional
        .Filename to save plot as EPS
    save_svg: str, optional
        .Filename to save plot as SVG
    save_pdf: str, optional
        .Filename to save plot as PDF
    save_jpeg: str, optional
        .Filename to save plot as JPEG
    display: bool, optional
        If True, then this function will display the plotted tplot variables. Necessary to make this optional
        so we can avoid it in a headless server environment.
    xsize: int, optional
        Size of the x-axis of the plot window in inches. Default is 10.
    ysize: int, optional
        Size of the y-axis of the plot window in inches. Default is 14.

    Returns:
    --------
    None

    Examples:
    ---------
    from pyspedas import mms_overview_plot
    mms_overview_plot(date='2023-03-23', probe='1', save_png="c:\\work\\mms\\summary_plots\mms_overview", display=False)

    Notes:
    ------
    It may need a long time to complete.

    Todo:
    -----
    1. Implement time_clip. Currently, it is not working correctly for spectra.
    2. Fix legends. The option "legend_location"="spedas", does not work correctly, it hides part of the legend, both on screen and in the saved PNG file.
    3. Implement thm_gen_multipngplot (6 hours, 2 hours).
    4. Fix problems when data is missing.
    """

    # If trange is not specified, use date and duration
    if (
        isinstance(trange, list)
        and len(trange) == 2
        and isinstance(trange[0], str)
        and isinstance(trange[1], str)
    ):
        t1 = time_double(trange[0])
        t2 = time_double(trange[1])
        trange = [t1, t2]
    else:
        t1 = time_double(date)
        t2 = t1 + float(duration) * 86400.0
        trange = [t1, t2]

    time_clip_value = False

    # Empty panel
    store_data("empty_panel", data={"x": trange, "y": [0, 0]})

    # Panel 1: AE index
    panel01 = "thg_idx_combined"
    if not skip_ae_idx:
        gmag(sites="idx", trange=trange, time_clip=time_clip_value)
        d1_1 = "thg_idx_ae"
        d1_2 = "thg_idx_uc_avg"

        store_data(panel01, data=[d1_1, d1_2])
        options(panel01, "name", "")
        options(panel01, "Color", ["black", "green"])
        options(panel01, "legend_names", ["Themis AE", "Kyoto proxy AE"])
        options(panel01, "ytitle", "AE Intex")
        options(panel01, "ysubtitle", "[nt]")
        options(panel01, "legend_location", "spedas")

    # Panel 2: FGM magnetic field mms_fpi_density
    panel02 = "mms" + probe + "_fgm_b_gsm_srvy_l2_bvec"
    fgm(trange=trange, probe=probe, data_rate="srvy", time_clip=time_clip_value)
    options(panel02, "ytitle", "FGM")
    options(panel02, "ysubtitle", "[nT]")
    options(panel02, "legend_location", "spedas")

    # Panel 3: EPD proton flux mms_epd_eis_extof_proton_flux_omni epd_eis_srvy_l2_extof_proton_flux_omni
    panel03 = "mms" + probe + "_epd_eis_srvy_l2_extof_proton_flux_omni"
    eis(
        trange=trange,
        probe=probe,
        datatype="extof",
        data_rate="srvy",
        time_clip=time_clip_value,
    )
    # options(panel03, "ytitle", "EPD\nEIS\nProton\nenergy")
    options(panel03, "ytitle", "EPD EIS\nProton energy")
    options(panel03, "ysubtitle", "[keV]")
    options(panel03, "legend_location", "spedas")

    # Panel 4: DIS Energy spectrogram mms_dis_energyspectr_omni_fast
    panel04 = "mms" + probe + "_dis_energyspectr_omni_fast"
    fpi(
        trange=trange,
        probe=probe,
        datatype=["des-moms", "dis-moms"],
        time_clip=time_clip_value,
    )
    options(panel04, "ytitle", "DIS\nenergy\nomni fast")
    options(panel04, "ysubtitle", "[eV]")
    options(panel04, "legend_location", "spedas")

    # Panel 5: Electron intensity mms_epd_feeps_srvy_l2_electron_intensity_omni
    panel05 = "mms" + probe + "_epd_feeps_srvy_l2_electron_intensity_omni"
    feeps(trange=trange, probe=probe, time_clip=time_clip_value)
    options(panel05, "ytitle", "Electron\nintensity\nomni")
    options(panel05, "ysubtitle", "[keV]")
    options(panel05, "legend_location", "spedas")

    # Panel 6: DES Energy spectrogram mms_des_energyspectr_omni_fast
    panel06 = "mms" + probe + "_des_energyspectr_omni_fast"
    options(panel06, "ytitle", "DES\nenergy\nomni fast")
    options(panel06, "ysubtitle", "[eV]")
    options(panel06, "legend_location", "spedas")

    # Panel 7: HPCA H- density mms_hpca_hplus_number_density
    panel07 = "mms" + probe + "_hpca_hplus_number_density"
    hpca(trange=trange, probe=probe, time_clip=time_clip_value)
    options(panel07, "ytitle", "HPCA\nH- density")
    options(panel07, "ysubtitle", "[cm-3]")
    options(panel07, "Color", ["blue"])
    options(panel07, "legend_location", "spedas")

    # Panel 8: EDP mms_edp_dce_dsl_fast_l2
    panel08 = "mms" + probe + "_edp_dce_dsl_fast_l2"
    edp(trange=trange, probe=probe, time_clip=time_clip_value)
    options(panel08, "ytitle", "EDP\nE Field")
    options(panel08, "ysubtitle", "[mV/m]")
    options(panel08, "legend_location", "spedas")

    # Panel 9: mms_fpi_density
    panel09 = "mms" + probe + "_fpi_density"
    d8_1 = "mms" + probe + "_dis_numberdensity_fast"
    d8_2 = "mms" + probe + "_des_numberdensity_fast"
    if tnames(d8_1) != [] and tnames(d8_2) != []:
        store_data(panel09, data=[d8_1, d8_2])
        options(panel09, "Color", ["green", "blue"])
        options(panel09, "legend_names", ["Ni, ions", "Ne, electrons"])
    elif tnames(d8_1) == [] and tnames(d8_2) == []:
        panel09 = "empty_panel"
    elif tnames(d8_1) == []:
        panel09 = d8_2
        options(panel09, "Color", ["blue"])
        options(panel09, "legend_names", ["Ne, electrons"])
    elif tnames(d8_2) == []:
        panel09 = d8_1
        options(panel09, "Color", ["green"])
        options(panel09, "legend_names", ["Ni, ions"])
    options(panel09, "name", "")
    options(panel09, "ytitle", "DIS")
    options(panel09, "ysubtitle", "[cm^-3]")
    options(panel09, "legend_location", "spedas")

    # Panel 10: HPCA H+ velocity mms_hpca_hplus_ion_bulk_velocity_GSM
    panel10 = "mms" + probe + "_hpca_hplus_ion_bulk_velocity_GSM"
    options(panel10, "ytitle", "HPCA\nH+ vel")
    options(panel10, "ysubtitle", "[km/s]")
    options(panel10, "legend_names", ["Vx(H^+)", "Vy(H^+)", "Vz(H^+)"])
    options(panel10, "legend_location", "spedas")

    # Panel 11: HPCA O- velocity mms_hpca_oplus_ion_bulk_velocity_GSM
    panel11 = "mms" + probe + "_hpca_oplus_ion_bulk_velocity_GSM"
    options(panel11, "ytitle", "HPCA\nO+ vel")
    options(panel11, "ysubtitle", "[km/s]")
    options(panel11, "legend_names", ["Vx(O^+)", "Vy(O^+)", "Vz(O^+)"])
    options(panel11, "legend_location", "spedas")

    # Save plot
    fname = None
    if isinstance(t1, str):
        t0 = time_string(time_double(t1), fmt="%Y_%m_%d")
    else:
        t0 = time_string(t1, fmt="%Y_%m_%d")

    # Full plot
    vars = [
        panel02,
        panel03,
        panel04,
        panel05,
        panel06,
        panel07,
        panel08,
        panel09,
        panel10,
        panel11,
    ]

    if not skip_ae_idx:
        vars.insert(0, panel01)

    # time_clip(vars, time_start=trange[0], time_end=trange[1])
    plot_title = "MMS-" + probe + " Overview (" + t0.replace("_", "-") + ")"
    tplot_options("title", plot_title)
    options(panel01, "name", plot_title)
    # tplot_options("wsize", [1400, 1000]) does not work
    tplot_options("show_all_axes", False)
    xsize = int(xsize)
    ysize = int(ysize)

    tplot(vars, save_png=save_png,save_eps=save_eps,save_svg=save_svg,save_pdf=save_pdf,save_jpeg=save_jpeg, display=display, xsize=xsize, ysize=ysize)
    logging.info("MMS overview plot completed.")



# Filtering Utilities
import enum
import numpy as np
from copy import deepcopy
from collections.abc import Iterable
from pyspedas.tplot_tools import tplot_names, data_quants, get_timespan, get_data, store_data

def filter_fields(tvars, dqflag = 0, keep = False, names_out = True):
    """
    This function filters on flagged values from PSP FIELDS magnetometer tplot 
    variables based on selected quality flags.  By default it filters OUT the 
    data where selected flags are set, but the keyword parameter "keep" can be 
    set True to keep just the data where the selected quality flags are set.
    
    For each TVAR passed in a new tplot variable is created with the filtered 
    data.  The new name is of the form:  <tvarname>_XXXXXX 
    where each 'XXX' is a 0 padded flag indicator sorted from lowest to highest.
    [If keep=True, the new name is of the form: <tvarname>_KXXXXXX]
 
    Flags 0 and -1 are special cases where the 'keep' flag is ignored

    So, psp.filter_fields('mag_RTN_1min_x',[4,16]) 
        - results in tvar named "mag_RTN_1min_x_004016"
        - tvar holds data where *NEITHER* flag 4 NOR flag 16 is set.

    And, psp.filter_fields('mag_RTN_1min_x',[4,16], keep=True) 
        - results in tvar named "mag_RTN_1min_x_K004016"
        - tvar holds data where *BOTH* flag 4 AND flag 16 is set.

    psp.filter_fields('mag_RTN_1min',0) 
        - results in tvar named "mag_RTN_1min_000"
        - tvar holds data with no set flags
    
    Valid only for variables of type:
        '...psp_fld_l2_mag_...' (full res, 1 minute, or 4 Sa per cycle)
        '...psp_fld_l2_rfs_...' (lfr and hfr) 

    Parameters
    ----------
        tvars: str or list of strings
            Tplot variable names, previously loaded

        dqflag: int or list of ints
            Elements indicate which of the data quality flags
            to filter on.             
            From the set {-1,0,1,2,4,8,16,32,64,128,256,512} 

            For keep=False (default)
            For flags >= 1: filters OUT wherever ANY of the selected flags are set.

            For keep=True
            For flags >= 1: returns data wherever ANY of the selected flags are set.

            Note: if using 0 or -1, no other flags should be selected
            -1: Keep cases with no set flags or only the 128 flag set 
            0: No set flags. (default)
            1: FIELDS antenna bias sweep
            2: PSP thruster firing
            4: SCM Calibration
            8: PSP rotations for MAG calibration (MAG rolls)
            16: FIELDS MAG calibration sequence
            32: SWEAP SPC in electron mode
            64: PSP Solar limb sensor (SLS) test
            128: PSP spacecraft is off umbra pointing  
            256: High frequency noise affecting RFS and TDS receivers. 
            512: Antennas driven towards the FIELDS power supply rails.   

        keep: bool
            If True, for dqflags >=1, return data where any of the selected flags 
            are set.  Otherwise, return data where none of the selected flags 
            are set.  (Default False)

        names_out: bool
            If set, return a list of the tplot variable names created 
            from this filter. Order corresponds to the input array of tvar
            names, so that tvar[i] filtered is in names_out[i]
            (Default True)

    Returns
    ----------
        None or List of tplot variables created if name_out=True (default)
    """

    new_names = [] if names_out else None

    # Check arguments
    if isinstance(tvars, str):
        tvars = [tvars]

    if not isinstance(dqflag, Iterable):
        dqflag = [dqflag]

    for flg in dqflag:
        assert flg in {-1,0,1,2,4,8,16,32,64,128,256,512}, "Bad FIELDS Quality Flag: {}".format(flg)

    if len(dqflag) > 1:
        assert 0 not in dqflag, 'DQFLAG of 0 must be set by itself'
        assert -1 not in dqflag, 'DQFLAG of -1 must be set by itself'

    dqflag.sort()

    # Reduce list to supported variables that are already loaded
    tvars = [x for x in tvars if x in tplot_names(quiet=True)]

    tvars = [x for x in tvars if 'fld_l2_mag_RTN' in x 
                              or 'fld_l2_mag_SC' in x 
                              or 'rfs_lfr' in x
                              or 'rfs_hfr' in x]
    if len(tvars) == 0:
        print("No valid variables for filtering")
        return new_names       

    # Create time specific quality flags as needed
    qf_target = []
    err_flg = []
    keep_var = []
    for tvar in tvars:

        qf_root = data_quants[tvar].qf_root
        if qf_root:
            keep_var.append(tvar)
            if ('rfs_hfr' in tvar) or ('rfs_lfr' in tvar):
                # Lots of possible epoch tags for the rfs data
                tag = '_' + data_quants[tvar].CDF['VATT']['DEPEND_0']
            elif '_1min' in tvar:
                tag = '_1min'
            elif '_4_Sa_per_Cyc' in tvar:
                tag = '_4_per_cycle'
            else:
                tag = '_hires'

            make_qf_var = False
            if qf_root + tag in data_quants.keys():
                # Variable exists.  Is it in the right time range?
                varTR = get_timespan(tvar)
                qfTR = get_timespan(tvar)
                if (qfTR[0] != varTR[0]) or (qfTR[1] != varTR[1]):
                    make_qf_var = True                
            else:
                make_qf_var = True

            if make_qf_var:
                err = _ff_match_qf_epochs(tvar, qf_root, tag)
                err_flg.append(err)
            else:
                err_flg.append(0)

            qf_target.append(qf_root + tag)
        else:
            print("Source quality flags missing. No filtered variabled created for: {}".format(tvar))
    
    # Remove cases where match epochs couldn't work
    valid_tn = [keep_var[i] for i in range(len(keep_var)) if err_flg[i] == 0]
    qf_target = [qf_target[i] for i in range(len(keep_var)) if err_flg[i] == 0]
    bad_tn = [keep_var[i] for i in range(len(keep_var)) if err_flg[i] == 1]
    if len(bad_tn) > 0:
        print("Error creating some matching time quality flags.")
        for tn in bad_tn:
            print("No filtered variabled created for: {}".format(tn))

    # Create the filtered variables
    if len(valid_tn) == 0:
        print("No valid variables for filtering\n")

    else:
        qfmap = dict()
        
        # Handle simpler cases, just the 0 or -1 flag
        if len(dqflag) == 1 and dqflag[0] in [0, -1]:
            # Get the quality flag arrays
            for qfname in qf_target:
                if qfname not in qfmap:
                    d = get_data(qfname)
                    qfmap[qfname] = d.y

            for i, tn in enumerate(valid_tn):
                if dqflag[0] == 0:
                    suffix = '_000'
                    mask = qfmap[qf_target[i]] != 0
                else:
                    suffix = '_0-1'
                    mask = (qfmap[qf_target[i]] != 0) & (qfmap[qf_target[i]] != 128)
                new_attrs = deepcopy(get_data(tn, metadata=True))
                if 'CDF' in new_attrs:
                    _ = new_attrs.pop('CDF')
                new_attrs['qf_root'] = None
                new_attrs['plot_options']['yaxis_opt']['axis_label'] += '\n(flt: {})'.format(suffix)  # TODO: how to increase left margin programmatically so this can be a new line?! (default matplotlib is .1 normal)
                
                data = deepcopy(get_data(tn))
                if issubclass(data.y.dtype.type, np.integer):
                    new_data = {'x':data.times, 'y':data.y.astype(float)}
                else:
                    new_data = {'x':data.times, 'y':data.y}

                try:
                    new_data['v'] = data.v
                except:
                    pass
   
                new_data['y'][mask] = np.nan

                store_data(tn+suffix, new_data, attr_dict=new_attrs)
                if names_out:
                    new_names.append(tn+suffix)

        # Handle multiple flags
        else:
            # Get quality flag arrays and flagged bits
            for qfname in qf_target:
                if qfname not in qfmap:
                    d = get_data(qfname)
                    qfmap[qfname] = dict()
                    qfmap[qfname]['dqf'] = d.y
                    qfmap[qfname]['dqfbits'] = _ff_unpackbits(d.y)

            suffix = '_K' if keep else '_'
            mybits = _ff_unpackbits(np.uint32(0))
            for flg in dqflag:
                suffix += '{:03}'.format(flg)
                mybits += _ff_unpackbits(np.uint32(flg))

            for i, tn in enumerate(valid_tn):
                new_attrs = deepcopy(get_data(tn, metadata=True))
                if 'CDF' in new_attrs:
                    _ = new_attrs.pop('CDF')
                new_attrs['qf_root'] = None
                new_attrs['plot_options']['yaxis_opt']['axis_label'] += '\n(flt: {})'.format(suffix) # TODO: how to increase left margin programmatically so this can be a new line and visisble by default?! (default matplotlib is .1 normal)
                
                data = deepcopy(get_data(tn))
                if issubclass(data.y.dtype.type, np.integer):
                    new_data = {'x':data.times, 'y':data.y.astype(float)}
                else:
                    new_data = {'x':data.times, 'y':data.y}

                try:
                    new_data['v'] = data.v
                except:
                    pass

                rem_mask = np.full(len(new_data['y']), False)
                for j in range(len(mybits)):
                    if mybits[j] == 1:
                        mask = qfmap[qf_target[i]]['dqfbits'][:, j] == 1
                        rem_mask[mask] = True

                if keep:
                    rem_mask = ~rem_mask
                new_data['y'][rem_mask] = np.nan

                if rem_mask.all():
                    #Handle completely empty data case.    
                    print("{}: (Skipped) No data matches this filter".format(tn))
                else:
                    store_data(tn+suffix, new_data, attr_dict=new_attrs)
                    if names_out:
                        new_names.append(tn+suffix)
    return new_names


def _ff_unpackbits(x):
    """
    Like np.unpackbits but works for all uint dtypes (not just uint8)
    Adapated from https://stackoverflow.com/a/51509307
    """
    if np.issubdtype(x.dtype, np.floating):
        raise ValueError("numpy data type needs to be int-like")
    num_bits = 8 * x.dtype.itemsize

    xshape = list(x.shape)
    x = x.reshape([-1, 1])
    mask = 2**np.arange(num_bits, dtype=x.dtype).reshape([1, num_bits])
    return (x & mask).astype(bool).astype(int).reshape(xshape + [num_bits])


def _ff_match_qf_epochs(target, source, tag):
    """
    Helper to psp.filter_fields routine.  Extends the quality_flag variable to
    match the time resolution target.  Results valid only if the source epochs
    and target epochs cover the same time period.
    
    Points in the target epoch falling before the source epoch start time will 
    contain the quality flag of the earliest entry in the quality flag variable.
    
    Points in the target epoch falling after the source epoch end time will 
    contain the quality flag of the last entry in the quality flag variable.

    Parameters
    ----------
        target: str
            Tplot variable name with the target epoch times

        source: str
            Tplot variable name with the source epoch times (ie: root quality flag variable)

        tag: str
            Tplot variable name suffix for new quality flag variable
    
    Returns
    ----------
        error: int
            Error status
            0 for nominal execution
            1 if the source and target epoch ranges are completely disjoint
                or do not overlap by at least 90% of the target time range.
                No new tplot variables are created in this instance.    
    """
    
    qfd = get_data(source)
    src = qfd.times
    tar = get_data(target).times.copy()

    # Check for valid ranges
    if (tar[-1] < src[0]) or (tar[0] > src[-1]):
        print("Epoch ranges are disjoint.")
        return 1
    else:
        tar_dur = tar[-1] - tar[0]
        src_overlap = src[(src >= tar[0]) & (src <= tar[-1])]
        overlap_dur = src_overlap[-1] - src_overlap[0]
        if overlap_dur/tar_dur < 0.9:
            print("Bad epoch ranges.")
            return 1

    newY = np.full_like(tar, qfd.y[0], dtype=qfd.y.dtype)
    for i in range(len(src) - 1):
        mask = (tar >= src[i]) & (tar < src[i+1])
        newY[mask] = qfd.y[i].copy()
    mask = tar >= src[-1]
    newY[mask] = qfd.y[-1].copy()

    newData = {'x':tar, 'y':newY}
    store_data(source+tag, newData)
    return 0



from .load import load

# This routine was originally in image/__init__.py, until being moved to its own file.
# Please refer to __init__.py if you need to see the revision history before it was moved.

def euv(
    trange=["2004-11-5", "2004-11-6"],
    datatype="k0",
    suffix='',
    prefix='',
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Loads IMAGE EUV data.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest with the format ['YYYY-MM-DD','YYYY-MM-DD'] or
        to specify more or less than a day ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Defaults to ['2004-11-5', '2004-11-6'].
    datatype : str, optional
        Data type. Defaults to 'k0'.
    suffix : str, optional
        The tplot variable names will be given this suffix. Defaults to ''.
    prefix : str, optional
        The tplot variable names will be given this prefix. Defaults to ''.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        By default, only loads in data with a "VAR_TYPE" attribute of "data". Defaults to False.
    varformat : str, optional
        The file variable formats to load into tplot. Wildcard character "*" is accepted.
        By default, all variables are loaded in. Defaults to None.
    varnames : list of str, optional
        List of variable names to load. If not specified, all data variables are loaded. Defaults to [].
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables. Defaults to False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables. Defaults to False.
    no_update : bool, optional
        If set, only load data from your local cache. Defaults to False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword. Defaults to False.
    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------
    list
        List of tplot variables created.
    """
    return load(
        instrument="euv",
        trange=trange,
        datatype=datatype,
        suffix=suffix,
        prefix=prefix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        downloadonly=downloadonly,
        notplot=notplot,
        time_clip=time_clip,
        no_update=no_update,
        force_download=force_download,
    )



from .load import load
from pyspedas.utilities.datasets import find_datasets

# This routine was moved out of __init__.py. Please see that file for previous revision history.


def threedp(trange=['1999-11-5', '1999-11-6'],
        datatype='3dp_emfits_e0',
        prefix='',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        force_download=False,
        notplot=False,
        no_update=False,
        berkeley=False,
        time_clip=False,
        addmaster=True):
    """
    Load WIND 3DP data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['1999-11-5', '1999-11-6']

        datatype: str
            Data type; Valid options: '3dp_ehpd', '3dp_ehsp', '3dp_elm2', '3dp_elpd', '3dp_elsp', '3dp_em', '3dp_emfits_e0',
            '3dp_k0', '3dp_phsp', '3dp_plsp', '3dp_pm', '3dp_sfpd', '3dp_sfsp', '3dp_sopd', '3dp_sosp'
            Default: '3dp_emfits_e0'

        prefix: str
            The tplot variable names will be given this prefix.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: None (all variables will be loaded)

        varnames: list of str
            List of variable names to load (if empty list or not specified,
            all data variables are loaded)
            Default: [] (all variables will be loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables
            Default: False

        force_download: bool
            Set this flag to download the CDF files, even if the local copy is newer.
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        addmaster: bool
            If True, use the metadata from a master CDF at SPDF rather than the metadata in the data file
            Default: True

    Returns
    ----------
        List of str
            List of tplot variables created.

    Examples
    --------

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> vars = pyspedas.projects.wind.threedp(trange=['1999-11-5', '1999-11-6'],datatype='3dp_emfits_e0')
    >>> tplot(vars)
    """
    return load(instrument='3dp', berkeley=berkeley, addmaster=addmaster, trange=trange, datatype=datatype, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, force_download=force_download, notplot=notplot, time_clip=time_clip, no_update=no_update)




import os
import logging
from viresclient import ClientConfig
import configparser

CONFIG = {
    'local_data_dir': 'vires_data/',
    'remote_data_dir': 'https://vires.services/ows',
    'access_token': ''
}

if os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir']  = os.environ['SPEDAS_DATA_DIR'] # self-defined, e.g. '/Users/cao/Documents/Data_N_Res/data_local'

if os.environ.get('VIRES_DATA_DIR'):
    CONFIG['local_data_dir']  = os.environ['VIRES_DATA_DIR'] # self-defined, e.g. '/Users/cao/Documents/Data_N_Res/data_local'

if os.environ.get('VIRES_TOKEN'):
    CONFIG['access_token'] = os.environ['VIRES_TOKEN']
else:
    homedir = os.path.expanduser('~')
    token_file = os.path.join(homedir, '.viresclient.ini')

    # try to read from token file
    try:
        config = configparser.ConfigParser()
        config.read(token_file)
        access_token = config.get('https://vires.services/ows','token')
        CONFIG['access_token'] = access_token
    except FileNotFoundError:
        logging.warning('Unable to load VIRES access token from VIRES_TOKEN environment variable or token file %s.',
                        token_file)
        logging.warning(
            'Visit https://viresclient.readthedocs.io/en/latest/access_token.html to learn how to set up an account and generate an access token.')
    except configparser.NoSectionError:
        logging.warning('Unable to load VIRES access token from VIRES_TOKEN environment variable or token file %s.', token_file)
        logging.warning('Visit https://viresclient.readthedocs.io/en/latest/access_token.html to learn how to set up an account and generate an access token.')



import logging
import re
import os
from . import download_files_utilities as utils
from .file_regex import maven_kp_l2_regex
import numpy as np
import collections

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def remove_inst_tag(df):
    """
    Remove the leading part of the column name that includes the instrument
    identifier for use in creating the parameter names for the toolkit.

    Parameters:
        A DataFrame produced from the insitu KP data
    Output:
        A new set of column names
    """

    newcol = []
    for i in df.columns:
        if len(i.split(".")) >= 2:
            j = i.split(".")
            newcol.append(".".join(j[1:]))

    return newcol


def get_latest_files_from_date_range(date1, date2):
    """
    Retrieves the latest files from a given date range.

    Parameters:
        date1 (datetime): The start date of the range.
        date2 (datetime): The end date of the range.

    Returns:
        list: A list of file paths representing the latest files within the date range.
    """

    from datetime import timedelta

    mvn_root_data_dir = utils.get_root_data_dir()
    sep = "/" if is_fsspec_uri(mvn_root_data_dir) else os.path.sep
    maven_data_dir = sep.join([
        mvn_root_data_dir, "maven", "data", "sci", "kp", "insitu"
    ])

    # Each file starts at midnight, so lets cut off the hours and just pay attention to the days
    date1 = date1.replace(hour=0, minute=0, second=0)
    date2 = date2.replace(hour=0, minute=0, second=0) + timedelta(days=1)

    time_spanned = date2 - date1
    num_days = time_spanned.days

    filenames = []
    kp_regex, l2_regex = maven_kp_l2_regex()

    for i in range(num_days):
        current_date = date1 + timedelta(days=i)
        year = str(current_date.year)
        month = str("%02d" % current_date.month)
        day = str("%02d" % current_date.day)
        full_path = sep.join([maven_data_dir, year, month])
        listdir = []
        if is_fsspec_uri(full_path):
            protocol, path = maven_data_dir.split("://")
            fs = fsspec.filesystem(protocol)

            exists = fs.exists(full_path)
            if exists:
                listdir = fs.listdir(full_path, detail=False)
                # fsspec alternative to os.path.basename
                listdir = [f.rstrip("/").split("/")[-1] for f in listdir]
        else:
            exists = os.path.exists(full_path)
            if exists: listdir = os.listdir(full_path)
        if exists:
            # Grab only the most recent version/revision of regular and crustal insitu files for each
            # day
            insitu = {}
            c_insitu = {}
            for f in listdir:
                # logging.warning(f)
                if kp_regex.match(f).group("day") == day and not kp_regex.match(
                    f
                ).group("description"):
                    v = kp_regex.match(f).group("version")
                    r = kp_regex.match(f).group("revision")
                    insitu[f] = [v, r]
                elif (
                    kp_regex.match(f).group("day") == day
                    and kp_regex.match(f).group("description") == "_crustal"
                ):
                    v = kp_regex.match(f).group("version")
                    r = kp_regex.match(f).group("revision")
                    c_insitu[f] = [v, r]

            if insitu:
                # Get max version
                insitu_file = max(insitu.keys(), key=(lambda k: insitu[k][0]))
                max_v = re.search(r"v\d{2}", insitu_file).group(0)
                # Get max revision
                max_r = max(
                    [re.search(r"r\d{2}", k).group(0) for k in insitu if max_v in k]
                )
                # Get most recent insitu file (based on max version, and then max revision values)
                most_recent_insitu = [
                    f for f in insitu.keys() if max_r in f and max_v in f
                ]
                filenames.append(sep.join([full_path, most_recent_insitu[0]]))

            if c_insitu:
                # Get max version
                c_insitu_file = max(c_insitu.keys(), key=(lambda k: c_insitu[k][0]))
                c_max_v = re.search(r"v\d{2}", c_insitu_file).group(0)
                # Get max revision
                c_max_r = max(
                    [re.search(r"r\d{2}", k).group(0) for k in c_insitu if c_max_v in k]
                )
                # Get most recent insitu file (based on max version, and then max revision values)
                most_recent_c_insitu = [
                    f for f in c_insitu.keys() if c_max_r in f and c_max_v in f
                ]
                filenames.append(sep.join([full_path, most_recent_c_insitu[0]]))

    filenames = sorted(filenames)
    return filenames


def get_latest_iuvs_files_from_date_range(date1, date2):
    """
    Get the latest IUVS files from a given date range.

    Parameters:
        date1 (datetime): The start date of the range.
        date2 (datetime): The end date of the range.

    Returns:
        list: A sorted list of file paths for the latest IUVS files within the date range.
    """

    from datetime import timedelta

    kp_regex, l2_regex = maven_kp_l2_regex()
    mvn_root_data_dir = utils.get_root_data_dir()
    sep = "/" if is_fsspec_uri(mvn_root_data_dir) else os.path.sep
    maven_data_dir = sep.join([
        mvn_root_data_dir, "maven", "data", "sci", "kp", "iuvs"
    ])

    # Each file starts at midnight, so lets cut off the hours and just pay attention to the days
    date1 = date1.replace(hour=0, minute=0, second=0)
    date2 = date2.replace(day=date2.day, hour=0, minute=0, second=0) + timedelta(days=1)

    time_spanned = date2 - date1
    num_days = time_spanned.days

    files_to_return = []
    for i in range(num_days):
        current_date = date1 + timedelta(days=i)
        year = str(current_date.year)
        month = str("%02d" % current_date.month)
        day = str("%02d" % current_date.day)
        # TODO: Fix the path after we fix the iuvs regex
        full_path = sep.join([maven_data_dir,"occ-", "02"])
        listdir = []
        if is_fsspec_uri(full_path):
            protocol, path = maven_data_dir.split("://")
            fs = fsspec.filesystem(protocol)

            exists = fs.exists(full_path)
            if exists:
                listdir = fs.listdir(full_path, detail=False)
                # fsspec alternative to os.path.basename
                listdir = [f.rstrip("/").split("/")[-1] for f in listdir]
        else:
            exists = os.path.exists(full_path)
            if exists: listdir = os.listdir(full_path)
        if exists:
            basenames = []
            # Obtain a list of all the basenames for the day
            for f in listdir:
                if kp_regex.match(f).group("day") == day:
                    description = kp_regex.match(f).group("description")
                    year = kp_regex.match(f).group("year")
                    month = kp_regex.match(f).group("month")
                    day = kp_regex.match(f).group("day")
                    time = kp_regex.match(f).group("time")
                    seq = ("mvn", "kp", "iuvs" + description, year + month + day + time)
                    basenames.append("_".join(seq))

            basenames = list(set(basenames))

            for bn in basenames:
                version = 0
                revision = 0
                for f in listdir:
                    description = kp_regex.match(f).group("description")
                    year = kp_regex.match(f).group("year")
                    month = kp_regex.match(f).group("month")
                    day = kp_regex.match(f).group("day")
                    time = kp_regex.match(f).group("time")
                    seq = ("mvn", "kp", "iuvs" + description, year + month + day + time)
                    if bn == "_".join(seq):
                        v = kp_regex.match(f).group("version")
                        if int(v) > int(version):
                            version = v

                for f in listdir:
                    description = kp_regex.match(f).group("description")
                    year = kp_regex.match(f).group("year")
                    month = kp_regex.match(f).group("month")
                    day = kp_regex.match(f).group("day")
                    time = kp_regex.match(f).group("time")
                    file_version = kp_regex.match(f).group("version")
                    seq = ("mvn", "kp", "iuvs" + description, year + month + day + time)
                    if bn == "_".join(seq) and file_version == version:
                        r = kp_regex.match(f).group("revision")
                        if int(r) > int(revision):
                            revision = r
                if int(version) > 0:
                    seq = (bn, "v" + str(version), "r" + str(revision) + ".tab")
                    files_to_return.append(sep.join([full_path, "_".join(seq)]))
    files_to_return = sorted(files_to_return)
    return files_to_return


def get_l2_files_from_date(date1, instrument):
    """
    Retrieves a list of Level 2 files for a given date and instrument.

    Parameters:
        date1 (datetime.datetime): The date for which to retrieve the files.
        instrument (str): The instrument name.

    Returns:
        list: A sorted list of file paths for the Level 2 files.
    """

    kp_regex, l2_regex = maven_kp_l2_regex()
    mvn_root_data_dir = utils.get_root_data_dir()
    sep = "/" if is_fsspec_uri(mvn_root_data_dir) else os.path.sep
    maven_data_dir = sep.join([mvn_root_data_dir, "maven", "data", "sci", instrument, "l2"])

    # Each file starts at midnight, so lets cut off the hours and just pay attention to the days
    date1 = date1.replace(hour=0, minute=0, second=0)

    filenames = []

    year = str(date1.year)
    month = str("%02d" % date1.month)
    day = str("%02d" % date1.day)
    full_path = sep.join([maven_data_dir, year, month])
    listdir = []
    if is_fsspec_uri(mvn_root_data_dir):
        protocol, path = mvn_root_data_dir.split("://")
        fs = fsspec.filesystem(protocol)

        exists = fs.exists(full_path)
        if exists:
            listdir = fs.listdir(full_path, detail=False)
            # fsspec alternative to os.path.basename
            listdir = [f.rstrip("/").split("/")[-1] for f in listdir]
    else:
        exists = os.path.exists(full_path)
        if exists:
            listdir = os.listdir(full_path)

    if exists:
        for f in listdir:
            if l2_regex.match(f).group("day") == day:
                filenames.append(sep.join([full_path, f]))

    filenames = sorted(filenames)
    return filenames


def get_header_info(filename):
    """
    Extracts header information from a file.

    Parameters:
    filename (str): The path to the file.

    Returns:
    tuple: A tuple containing the names list and inst list.
    """
    # Determine number of header lines
    nheader = 0
    if is_fsspec_uri(filename):
        protocol, path = filename.split("://")
        fs = fsspec.filesystem(protocol)
        fo = fs.open(filename, "rt")
    else:
        fo = open(filename)

    with fo as f:
        for line in f:
            if line.startswith("#"):
                nheader += 1

    # Parse the header (still needs special case work)
    read_param_list = False
    start_temp = False
    index_list = []
    if is_fsspec_uri(filename):
        protocol, path = filename.split("://")
        fs = fsspec.filesystem(protocol)
        fo = fs.open(filename, "rt")
    else:
        fo = open(filename)
    with fo as fin:
        icol = -2  # Counting header lines detailing column names
        iname = 1  # for counting seven lines with name info
        ncol = -1  # Dummy value to allow reading of early headerlines?
        col_regex = r"#\s(.{16}){%3d}" % ncol  # needed for column names
        crustal = False
        if "crustal" in filename:
            crustal = True
        for iline in range(nheader):
            line = fin.readline()
            # Define the proper indices change depending on the file type and row
            i = [2, 2, 1] if crustal else [1, 1, 1]
            if re.search("Number of parameter columns", line):
                ncol = int(re.split(r"\s{3}", line)[i[0]])
                # needed for column names
                col_regex = (
                    r"#\s(.{16}){%2d}" % ncol if crustal else r"#\s(.{16}){%3d}" % ncol
                )
            elif re.search("Line on which data begins", line):
                nhead_test = int(re.split(r"\s{3}", line)[i[1]]) - 1
            elif re.search("Number of lines", line):
                ndata = int(re.split(r"\s{3}", line)[i[2]])
            elif re.search("PARAMETER", line):
                read_param_list = True
                param_head = iline
            elif read_param_list:
                icol += 1
                if icol > ncol:
                    read_param_list = False
            elif re.match(col_regex, line):
                # OK, verified match now get the values
                temp = re.findall("(.{16})", line[3:])
                if temp[0] == "               1":
                    start_temp = True
                if start_temp:
                    # Crustal files do not have as much variable info as other insitu files, need
                    # to modify the lines below
                    if crustal:
                        if iname == 1:
                            index = temp
                        elif iname == 2:
                            obs1 = temp
                        elif iname == 3:
                            obs2 = temp
                        elif iname == 4:
                            unit = temp
                            # crustal files don't come with this field
                            # throwing it in here for consistency with other insitu files
                            inst = ["     MODELED_MAG"] * 13
                        else:
                            logging.warning(
                                "More lines in data descriptor than expected."
                            )
                            logging.warning("Line %d" % iline)
                    else:
                        if iname == 1:
                            index = temp
                        elif iname == 2:
                            obs1 = temp
                        elif iname == 3:
                            obs2 = temp
                        elif iname == 4:
                            obs3 = temp
                        elif iname == 5:
                            inst = temp
                        elif iname == 6:
                            unit = temp
                        elif iname == 7:
                            format_code = temp
                        else:
                            logging.warning(
                                "More lines in data descriptor than expected."
                            )
                            logging.warning("Line %d" % iline)
                    iname += 1
            else:
                pass

        # Generate the names list.
        # NB, there are special case redundancies in there
        # (e.g., LPW: Electron Density Quality (min and max))
        # ****SWEA FLUX electron QUALITY *****
        first = True
        parallel = None
        names = []
        if crustal:
            for h, i, j in zip(inst, obs1, obs2):
                combo_name = (" ".join([i.strip(), j.strip()])).strip()
                # Add inst to names to avoid ambiguity
                # Will need to remove these after splitting
                names.append(".".join([h.strip(), combo_name]))
                names[0] = "Time"
        else:
            for h, i, j, k in zip(inst, obs1, obs2, obs3):
                combo_name = (" ".join([i.strip(), j.strip(), k.strip()])).strip()
                if re.match("^LPW$", h.strip()):
                    # Max and min error bars use same name in column
                    # SIS says first entry is min and second is max
                    if re.match("(Electron|Spacecraft)(.+)Quality", combo_name):
                        if first:
                            combo_name = combo_name + " Min"
                            first = False
                        else:
                            combo_name = combo_name + " Max"
                            first = True
                elif re.match("^SWEA$", h.strip()):
                    # electron flux qual flags do not indicate whether parallel or anti
                    # From context it is clear; but we need to specify in name
                    if re.match(".+Parallel.+", combo_name):
                        parallel = True
                    elif re.match(".+Anti-par", combo_name):
                        parallel = False
                    else:
                        pass
                    if re.match("Flux, e-(.+)Quality", combo_name):
                        if parallel:
                            p = re.compile("Flux, e- ")
                            combo_name = p.sub("Flux, e- Parallel ", combo_name)
                        else:
                            p = re.compile("Flux, e- ")
                            combo_name = p.sub("Flux, e- Anti-par ", combo_name)
                    if re.match("Electron eflux (.+)Quality", combo_name):
                        if parallel:
                            p = re.compile("Electron eflux ")
                            combo_name = p.sub("Electron eflux  Parallel ", combo_name)
                        else:
                            p = re.compile("Electron eflux ")
                            combo_name = p.sub("Electron eflux  Anti-par ", combo_name)
                # Add inst to names to avoid ambiguity
                # Will need to remove these after splitting
                names.append(".".join([h.strip(), combo_name]))
                names[0] = "Time"

    return names, inst

def maven_param_list():
    """
    Returns a dictionary of MAVEN parameters and their corresponding keys.

    Returns:
        dict: A dictionary mapping parameter names to their keys.
    """

    param_dict = {
        "Electron Density": "ELECTRON_DENSITY",
        "Electron Density Quality Min": "ELECTRON_DENSITY_QUAL_MIN",
        "Electron Density Quality Max": "ELECTRON_DENSITY_QUAL_MAX",
        "Electron Temperature": "ELECTRON_TEMPERATURE",
        "Electron Temperature Quality Min": "ELECTRON_TEMPERATURE_QUAL_MIN",
        "Electron Temperature Quality Max": "ELECTRON_TEMPERATURE_QUAL_MAX",
        "Spacecraft Potential": "SPACECRAFT_POTENTIAL",
        "Spacecraft Potential Quality Min": "SPACECRAFT_POTENTIAL_QUAL_MIN",
        "Spacecraft Potential Quality Max": "SPACECRAFT_POTENTIAL_QUAL_MAX",
        "E-field Power 2-100 Hz": "EWAVE_LOW_FREQ",
        "E-field 2-100 Hz Quality": "EWAVE_LOW_FREQ_QUAL_QUAL",
        "E-field Power 100-800 Hz": "EWAVE_MID_FREQ",
        "E-field 100-800 Hz Quality": "EWAVE_MID_FREQ_QUAL_QUAL",
        "E-field Power 0.8-1.0 Mhz": "EWAVE_HIGH_FREQ",
        "E-field 0.8-1.0 Mhz Quality": "EWAVE_HIGH_FREQ_QUAL_QUAL",
        "EUV Irradiance 0.1-7.0 nm": "IRRADIANCE_LOW",
        "Irradiance 0.1-7.0 nm Quality": "IRRADIANCE_LOW_QUAL",
        "EUV Irradiance 17-22 nm": "IRRADIANCE_MID",
        "Irradiance 17-22 nm Quality": "IRRADIANCE_MID_QUAL",
        "EUV Irradiance Lyman-alpha": "IRRADIANCE_LYMAN",
        "Irradiance Lyman-alpha Quality": "IRRADIANCE_LYMAN_QUAL",
        "Solar Wind Electron Density": "SOLAR_WIND_ELECTRON_DENSITY",
        "Solar Wind E- Density Quality": "SOLAR_WIND_ELECTRON_DENSITY_QUAL",
        "Solar Wind Electron Temperature": "SOLAR_WIND_ELECTRON_TEMPERATURE",
        "Solar Wind E- Temperature Quality": "SOLAR_WIND_ELECTRON_TEMPERATURE_QUAL",
        "Flux, e- Parallel (5-100 ev)": "ELECTRON_PARALLEL_FLUX_LOW",
        "Flux, e- Parallel (5-100 ev) Quality": "ELECTRON_PARALLEL_FLUX_LOW_QUAL",
        "Flux, e- Parallel (100-500 ev)": "ELECTRON_PARALLEL_FLUX_MID",
        "Flux, e- Parallel (100-500 ev) Quality": "ELECTRON_PARALLEL_FLUX_MID_QUAL",
        "Flux, e- Parallel (500-1000 ev)": "ELECTRON_PARALLEL_FLUX_HIGH",
        "Flux, e- Parallel (500-1000 ev) Quality": "ELECTRON_PARALLEL_FLUX_HIGH_QUAL",
        "Flux, e- Anti-par (5-100 ev)": "ELECTRON_ANTI_PARALLEL_FLUX_LOW",
        "Flux, e- Anti-par (5-100 ev) Quality": "ELECTRON_ANTI_PARALLEL_FLUX_LOW_QUAL",
        "Flux, e- Anti-par (100-500 ev)": "ELECTRON_ANTI_PARALLEL_FLUX_MID",
        "Flux, e- Anti-par (100-500 ev) Quality": "ELECTRON_ANTI_PARALLEL_FLUX_MID_QUAL",
        "Flux, e- Anti-par (500-1000 ev)": "ELECTRON_ANTI_PARALLEL_FLUX_HIGH",
        "Flux, e- Anti-par (500-1000 ev) Quality": "ELECTRON_ANTI_PARALLEL_FLUX_HIGH_QUAL",
        "Electron eflux Parallel (5-100 ev)": "ELECTRON_PARALLEL_FLUX_LOW",
        "Electron eflux Parallel (5-100 ev) Quality": "ELECTRON_PARALLEL_FLUX_LOW_QUAL",
        "Electron eflux Parallel (100-500 ev)": "ELECTRON_PARALLEL_FLUX_MID",
        "Electron eflux Parallel (100-500 ev) Quality": "ELECTRON_PARALLEL_FLUX_MID_QUAL",
        "Electron eflux Parallel (500-1000 ev)": "ELECTRON_PARALLEL_FLUX_HIGH",
        "Electron eflux Parallel (500-1000 ev) Quality": "ELECTRON_PARALLEL_FLUX_HIGH_QUAL",
        "Electron eflux Anti-par (5-100 ev)": "ELECTRON_ANTI_PARALLEL_FLUX_LOW",
        "Electron eflux Anti-par (5-100 ev) Quality": "ELECTRON_ANTI_PARALLEL_FLUX_LOW_QUAL",
        "Electron eflux Anti-par (100-500 ev)": "ELECTRON_ANTI_PARALLEL_FLUX_MID",
        "Electron eflux Anti-par (100-500 ev) Quality": "ELECTRON_ANTI_PARALLEL_FLUX_MID_QUAL",
        "Electron eflux Anti-par (500-1000 ev)": "ELECTRON_ANTI_PARALLEL_FLUX_HIGH",
        "Electron eflux Anti-par (500-1000 ev) Quality": "ELECTRON_ANTI_PARALLEL_FLUX_HIGH_QUAL",
        "Electron Spectrum Shape": "ELECTRON_SPECTRUM_SHAPE_PARAMETER",
        "Spectrum Shape Quality": "ELECTRON_SPECTRUM_SHAPE_PARAMETER_QUAL",
        "H+ Density": "HPLUS_DENSITY",
        "H+ Density Quality": "HPLUS_DENSITY_QUAL",
        "H+ Flow Velocity MSO X": "HPLUS_FLOW_VELOCITY_MSO_X",
        "H+ Flow MSO X Quality": "HPLUS_FLOW_VELOCITY_MSO_X_QUAL",
        "H+ Flow Velocity MSO Y": "HPLUS_FLOW_VELOCITY_MSO_Y",
        "H+ Flow MSO Y Quality": "HPLUS_FLOW_VELOCITY_MSO_Y_QUAL",
        "H+ Flow Velocity MSO Z": "HPLUS_FLOW_VELOCITY_MSO_Z",
        "H+ Flow MSO Z Quality": "HPLUS_FLOW_VELOCITY_MSO_Z_QUAL",
        "H+ Temperature": "HPLUS_TEMPERATURE",
        "H+ Temperature Quality": "HPLUS_TEMPERATURE_QUAL",
        "Solar Wind Dynamic Pressure": "SOLAR_WIND_DYNAMIC_PRESSURE",
        "Solar Wind Pressure Quality": "SOLAR_WIND_DYNAMIC_PRESSURE_QUAL",
        "STATIC Quality Flag": "STATIC_QUALITY_FLAG",
        "O+ Density": "OPLUS_DENSITY",
        "O+ Density Quality": "OPLUS_DENSITY_QUAL",
        "O2+ Density": "O2PLUS_DENSITY",
        "O2+ Density Quality": "O2PLUS_DENSITY_QUAL",
        "O+ Temperature": "OPLUS_TEMPERATURE",
        "O+ Temperature Quality": "OPLUS_TEMPERATURE_QUAL",
        "O2+ Temperature": "O2PLUS_TEMPERATURE",
        "O2+ Temperature Quality": "O2PLUS_TEMPERATURE_QUAL",
        "O2+ Flow Velocity MAVEN_APP X": "O2PLUS_FLOW_VELOCITY_MAVEN_APP_X",
        "O2+ Flow MAVEN_APP X Quality": "O2PLUS_FLOW_VELOCITY_MAVEN_APP_X_QUAL",
        "O2+ Flow Velocity MAVEN_APP Y": "O2PLUS_FLOW_VELOCITY_MAVEN_APP_Y",
        "O2+ Flow MAVEN_APP Y Quality": "O2PLUS_FLOW_VELOCITY_MAVEN_APP_Y_QUAL",
        "O2+ Flow Velocity MAVEN_APP Z": "O2PLUS_FLOW_VELOCITY_MAVEN_APP_Z",
        "O2+ Flow MAVEN_APP Z Quality": "O2PLUS_FLOW_VELOCITY_MAVEN_APP_Z_QUAL",
        "O2+ Flow Velocity MSO X": "O2PLUS_FLOW_VELOCITY_MSO_X",
        "O2+ Flow MSO X Quality": "O2PLUS_FLOW_VELOCITY_MSO_X_QUAL",
        "O2+ Flow Velocity MSO Y": "O2PLUS_FLOW_VELOCITY_MSO_Y",
        "O2+ Flow MSO Y Quality": "O2PLUS_FLOW_VELOCITY_MSO_Y_QUAL",
        "O2+ Flow Velocity MSO Z": "O2PLUS_FLOW_VELOCITY_MSO_Z",
        "O2+ Flow MSO Z Quality": "O2PLUS_FLOW_VELOCITY_MSO_Z_QUAL",
        "H+ Omni Flux": "HPLUS_OMNI_DIRECTIONAL_FLUX",
        "H+ Energy": "HPLUS_CHARACTERISTIC_ENERGY",
        "H+ Energy Quality": "HPLUS_CHARACTERISTIC_ENERGY_QUAL",
        "He++ Omni Flux": "HEPLUS_OMNI_DIRECTIONAL_FLUX",
        "He++ Energy": "HEPLUS_CHARACTERISTIC_ENERGY",
        "He++ Energy Quality": "HEPLUS_CHARACTERISTIC_ENERGY_QUAL",
        "O+ Omni Flux": "OPLUS_OMNI_DIRECTIONAL_FLUX",
        "O+ Energy": "OPLUS_CHARACTERISTIC_ENERGY",
        "O+ Energy Quality": "OPLUS_CHARACTERISTIC_ENERGY_QUAL",
        "O2+ Omni Flux": "O2PLUS_OMNI_DIRECTIONAL_FLUX",
        "O2+ Energy": "O2PLUS_CHARACTERISTIC_ENERGY",
        "O2+ Energy Quality": "O2PLUS_CHARACTERISTIC_ENERGY_QUAL",
        "H+ Direction MSO X": "HPLUS_CHARACTERISTIC_DIRECTION_MSO_X",
        "H+ Direction MSO Y": "HPLUS_CHARACTERISTIC_DIRECTION_MSO_Y",
        "H+ Direction MSO Z": "HPLUS_CHARACTERISTIC_DIRECTION_MSO_Z",
        "H+ Angular Width": "HPLUS_CHARACTERISTIC_ANGULAR_WIDTH",
        "H+ Width Quality": "HPLUS_CHARACTERISTIC_ANGULAR_WIDTH_QUAL",
        "Pickup Ion Direction MSO X": "DOMINANT_PICKUP_ION_CHARACTERISTIC_DIRECTION_MSO_X",
        "Pickup Ion Direction MSO Y": "DOMINANT_PICKUP_ION_CHARACTERISTIC_DIRECTION_MSO_Y",
        "Pickup Ion Direction MSO Z": "DOMINANT_PICKUP_ION_CHARACTERISTIC_DIRECTION_MSO_Z",
        "Pickup Ion Angular Width": "DOMINANT_PICKUP_ION_CHARACTERISTIC_ANGULAR_WIDTH",
        "Pickup Ion Width Quality": "DOMINANT_PICKUP_ION_CHARACTERISTIC_ANGULAR_WIDTH_QUAL",
        "Ion Flux FOV 1 F": "ION_ENERGY_FLUX__FOV_1_F",
        "Ion Flux FOV 1F Quality": "ION_ENERGY_FLUX__FOV_1_F_QUAL",
        "Ion Flux FOV 1 R": "ION_ENERGY_FLUX__FOV_1_R",
        "Ion Flux FOV 1R Quality": "ION_ENERGY_FLUX__FOV_1_R_QUAL",
        "Ion Flux FOV 2 F": "ION_ENERGY_FLUX__FOV_2_F",
        "Ion Flux FOV 2F Quality": "ION_ENERGY_FLUX__FOV_2_F_QUAL",
        "Ion Flux FOV 2 R": "ION_ENERGY_FLUX__FOV_2_R",
        "Ion Flux FOV 2R Quality": "ION_ENERGY_FLUX__FOV_2_R_QUAL",
        "Electron Flux FOV 1 F": "ELECTRON_ENERGY_FLUX___FOV_1_F",
        "Electron Flux FOV 1F Quality": "ELECTRON_ENERGY_FLUX___FOV_1_F_QUAL",
        "Electron Flux FOV 1 R": "ELECTRON_ENERGY_FLUX___FOV_1_R",
        "Electron Flux FOV 1R Quality": "ELECTRON_ENERGY_FLUX___FOV_1_R_QUAL",
        "Electron Flux FOV 2 F": "ELECTRON_ENERGY_FLUX___FOV_2_F",
        "Electron Flux FOV 2F Quality": "ELECTRON_ENERGY_FLUX___FOV_2_F_QUAL",
        "Electron Flux FOV 2 R": "ELECTRON_ENERGY_FLUX___FOV_2_R",
        "Electron Flux FOV 2R Quality": "ELECTRON_ENERGY_FLUX___FOV_2_R_QUAL",
        "Look Direction 1-F MSO X": "LOOK_DIRECTION_1_F_MSO_X",
        "Look Direction 1-F MSO Y": "LOOK_DIRECTION_1_F_MSO_Y",
        "Look Direction 1-F MSO Z": "LOOK_DIRECTION_1_F_MSO_Z",
        "Look Direction 1-R MSO X": "LOOK_DIRECTION_1_R_MSO_X",
        "Look Direction 1-R MSO Y": "LOOK_DIRECTION_1_R_MSO_Y",
        "Look Direction 1-R MSO Z": "LOOK_DIRECTION_1_R_MSO_Z",
        "Look Direction 2-F MSO X": "LOOK_DIRECTION_2_F_MSO_X",
        "Look Direction 2-F MSO Y": "LOOK_DIRECTION_2_F_MSO_Y",
        "Look Direction 2-F MSO Z": "LOOK_DIRECTION_2_F_MSO_Z",
        "Look Direction 2-R MSO X": "LOOK_DIRECTION_2_R_MSO_X",
        "Look Direction 2-R MSO Y": "LOOK_DIRECTION_2_R_MSO_Y",
        "Look Direction 2-R MSO Z": "LOOK_DIRECTION_2_R_MSO_Z",
        "Magnetic Field MSO X": "MSO_X",
        "Magnetic MSO X Quality": "MSO_X_QUAL",
        "Magnetic Field MSO Y": "MSO_Y",
        "Magnetic MSO Y Quality": "MSO_Y_QUAL",
        "Magnetic Field MSO Z": "MSO_Z",
        "Magnetic MSO Z Quality": "MSO_Z_QUAL",
        "Magnetic Field GEO X": "GEO_X",
        "Magnetic GEO X Quality": "GEO_X_QUAL",
        "Magnetic Field GEO Y": "GEO_Y",
        "Magnetic GEO Y Quality": "GEO_Y_QUAL",
        "Magnetic Field GEO Z": "GEO_Z",
        "Magnetic GEO Z Quality": "GEO_Z_QUAL",
        "Magnetic Field RMS Dev": "RMS_DEVIATION",
        "Magnetic RMS Quality": "RMS_DEVIATION_QUAL",
        "Density He": "HE_DENSITY",
        "Density He Precision": "HE_DENSITY_PRECISION",
        "Density He Quality": "HE_DENSITY_QUAL",
        "Density O": "O_DENSITY",
        "Density O Precision": "O_DENSITY_PRECISION",
        "Density O Quality": "O_DENSITY_QUAL",
        "Density CO": "CO_DENSITY",
        "Density CO Precision": "CO_DENSITY_PRECISION",
        "Density CO Quality": "CO_DENSITY_QUAL",
        "Density N2": "N2_DENSITY",
        "Density N2 Precision": "N2_DENSITY_PRECISION",
        "Density N2 Quality": "N2_DENSITY_QUAL",
        "Density NO": "NO_DENSITY",
        "Density NO Precision": "NO_DENSITY_PRECISION",
        "Density NO Quality": "NO_DENSITY_QUAL",
        "Density Ar": "AR_DENSITY",
        "Density Ar Precision": "AR_DENSITY_PRECISION",
        "Density Ar Quality": "AR_DENSITY_QUAL",
        "Density CO2": "CO2_DENSITY",
        "Density CO2 Precision": "CO2_DENSITY_PRECISION",
        "Density CO2 Quality": "CO2_DENSITY_QUAL",
        "Density 32+": "O2PLUS_DENSITY",
        "Density 32+ Precision": "O2PLUS_DENSITY_PRECISION",
        "Density 32+ Quality": "O2PLUS_DENSITY_QUAL",
        "Density 44+": "CO2PLUS_DENSITY",
        "Density 44+ Precision": "CO2PLUS_DENSITY_PRECISION",
        "Density 44+ Quality": "CO2PLUS_DENSITY_QUAL",
        "Density 30+": "NOPLUS_DENSITY",
        "Density 30+ Precision": "NOPLUS_DENSITY_PRECISION",
        "Density 30+ Quality": "NOPLUS_DENSITY_QUAL",
        "Density 16+": "OPLUS_DENSITY",
        "Density 16+ Precision": "OPLUS_DENSITY_PRECISION",
        "Density 16+ Quality": "OPLUS_DENSITY_QUAL",
        "Density 28+": "CO2PLUS_N2PLUS_DENSITY",
        "Density 28+ Precision": "CO2PLUS_N2PLUS_DENSITY_PRECISION",
        "Density 28+ Quality": "CO2PLUS_N2PLUS_DENSITY_QUAL",
        "Density 12+": "CPLUS_DENSITY",
        "Density 12+ Precision": "CPLUS_DENSITY_PRECISION",
        "Density 12+ Quality": "CPLUS_DENSITY_QUAL",
        "Density 17+": "OHPLUS_DENSITY",
        "Density 17+ Precision": "OHPLUS_DENSITY_PRECISION",
        "Density 17+ Quality": "OHPLUS_DENSITY_QUAL",
        "Density 14+": "NPLUS_DENSITY",
        "Density 14+ Precision": "NPLUS_DENSITY_PRECISION",
        "Density 14+ Quality": "NPLUS_DENSITY_QUAL",
        "APP Attitude GEO X": "ATTITUDE_GEO_X",
        "APP Attitude GEO Y": "ATTITUDE_GEO_Y",
        "APP Attitude GEO Z": "ATTITUDE_GEO_Z",
        "APP Attitude MSO X": "ATTITUDE_MSO_X",
        "APP Attitude MSO Y": "ATTITUDE_MSO_Y",
        "APP Attitude MSO Z": "ATTITUDE_MSO_Z",
        "Spacecraft GEO X": "GEO_X",
        "Spacecraft GEO Y": "GEO_Y",
        "Spacecraft GEO Z": "GEO_Z",
        "Spacecraft MSO X": "MSO_X",
        "Spacecraft MSO Y": "MSO_Y",
        "Spacecraft MSO Z": "MSO_Z",
        "Spacecraft GEO Longitude": "SUB_SC_LONGITUDE",
        "Spacecraft GEO Latitude": "SUB_SC_LATITUDE",
        "Spacecraft Solar Zenith Angle": "SZA",
        "Spacecraft Local Time": "LOCAL_TIME",
        "Spacecraft Altitude Aeroid": "ALTITUDE",
        "Spacecraft Attitude GEO X": "ATTITUDE_GEO_X",
        "Spacecraft Attitude GEO Y": "ATTITUDE_GEO_Y",
        "Spacecraft Attitude GEO Z": "ATTITUDE_GEO_Z",
        "Spacecraft Attitude MSO X": "ATTITUDE_MSO_X",
        "Spacecraft Attitude MSO Y": "ATTITUDE_MSO_Y",
        "Spacecraft Attitude MSO Z": "ATTITUDE_MSO_Z",
        "Mars Season (Ls)": "MARS_SEASON",
        "Mars-Sun Distance": "MARS_SUN_DISTANCE",
        "Subsolar Point GEO Longitude": "SUBSOLAR_POINT_GEO_LONGITUDE",
        "Subsolar Point GEO Latitude": "SUBSOLAR_POINT_GEO_LATITUDE",
        "Sub-Mars Point on the Sun Longitude": "SUBMARS_POINT_SOLAR_LONGITUDE",
        "Sub-Mars Point on the Sun Latitude": "SUBMARS_POINT_SOLAR_LATITUDE",
        "Rot matrix MARS -> MSO Row 1, Col 1": "T11",
        "Rot matrix MARS -> MSO Row 1, Col 2": "T12",
        "Rot matrix MARS -> MSO Row 1, Col 3": "T13",
        "Rot matrix MARS -> MSO Row 2, Col 1": "T21",
        "Rot matrix MARS -> MSO Row 2, Col 2": "T22",
        "Rot matrix MARS -> MSO Row 2, Col 3": "T23",
        "Rot matrix MARS -> MSO Row 3, Col 1": "T31",
        "Rot matrix MARS -> MSO Row 3, Col 2": "T32",
        "Rot matrix MARS -> MSO Row 3, Col 3": "T33",
        "Rot matrix SPCCRFT -> MSO Row 1, Col 1": "SPACECRAFT_T11",
        "Rot matrix SPCCRFT -> MSO Row 1, Col 2": "SPACECRAFT_T12",
        "Rot matrix SPCCRFT -> MSO Row 1, Col 3": "SPACECRAFT_T13",
        "Rot matrix SPCCRFT -> MSO Row 2, Col 1": "SPACECRAFT_T21",
        "Rot matrix SPCCRFT -> MSO Row 2, Col 2": "SPACECRAFT_T22",
        "Rot matrix SPCCRFT -> MSO Row 2, Col 3": "SPACECRAFT_T23",
        "Rot matrix SPCCRFT -> MSO Row 3, Col 1": "SPACECRAFT_T31",
        "Rot matrix SPCCRFT -> MSO Row 3, Col 2": "SPACECRAFT_T32",
        "Rot matrix SPCCRFT -> MSO Row 3, Col 3": "SPACECRAFT_T33",
    }
    return param_dict



import pyspedas
import numpy as np
from scipy.interpolate import interp1d
from pyspedas.tplot_tools.time_double import time_double
from pyspedas.projects.themis.analysis.scpot2dens import interpolate_to_scptime

# Nishimura version
def scpot2dens_nishimura(scpot, scptime, vth, vth_time, dens_i, dens_i_time, pos_gsm, pos_gsm_time, probe='a', ion_density_threshold=40, no_interp=False):
    """
    Calculates spacecraft potential derived density, using Nishimura formula, as in the IDL SPEDAS SCPOT2DENS_OPT_N

    This function processes spacecraft potential data (`scpot`), alongside interpolated electron thermal velocity (`Vth`),
    ion density (`dens_i`) data and spacecraft position (`pos`), to calculate electron and ion densities at specific
    time points (`scptime`). It interpolates all data to the `scpot` time base. If no_interp is set, only the position 
    data is interpolated. Ion density is used to flag bad data; the calculation is expected to fail for high Ion density (Gt 40cm^(-3))

    Parameters
    ----------
    scpot : array_like
        Spacecraft potential data, expected as a 1-D array. (call pyspedas.projects.themis.esa, DATATYPE = 'peer_sc_pot')
    scptime : array_like
        The time base of `scpot`.
    vth : array_like
        Electron thermal velocity data, expected as a 1-D array. (call pyspedas.projects.themis.esa, DATATYPE = 'peer_vthermal')
    vth_time : array_like
        The time base of `vth`.
    dens_i : array_like
        Ion density data, expected as a 1-D array. (call pyspedas.projects.themis.esa, DATATYPE = 'peir_density')
    dens_i_time : array_like
        The time base for `dens_i`.
    pos_gsm: array_like
        Spacecraft position in GSM (or any earth-centered coordinates.).
    pos_gsm_time: array_like
        The time base of `pos_gsm`
    probe : int or str, default='a'
        Identifier for the probe being used for measurements. Valid values are 'a', 'b', 'c', 'd', 'e'.
    ion_density_threshold: Scalar, int or float, default=40
        Threshold for ion density, above which, data is set to 'NaN'
    no_interp : Boolean
        If True, then all of the probe ESA arrays are the same (typical for THEMIS ESA data), no need to interpolate times.
        Position data is always interpolated

    Returns
    -------
    dens : ndarray
        The calculated densities at `scpot` time points. Returns None if the probe is invalid, if `dens_i`
        is not present or do not have at least two elements (indicating insufficient data for interpolation, 
        for no_interp=False)

    Notes
    -----
    - Incorporates probe-specific adjustments
    - Invalid inputs or insufficient data lead to an early termination of the function with None as return.

    Examples
    --------
    Given a specific time range, the following example demonstrates how to load THEMIS spacecraft data for probe 'a',
    extract density, spacecraft potential, and velocity data, then calculate the plasma density from spacecraft potential
    using the `scpot2dens` function.



    >>> import pyspedas
    >>> from pyspedas.projects.themis import scpot2dens_nishimura
    >>> from pyspedas import get_data
    >>>
    >>> # Define the time range for which to load the data
    >>> trange = ['2017-07-20', '2017-07-20']
    >>>
    >>> # Load THEMIS ESA data for probe 'a' within the specified time range
    >>> pyspedas.projects.themis.esa(trange=trange,
    >>>                     probe='a',
    >>>                     varnames=['tha_peir_density', 'tha_peer_sc_pot', 'tha_peer_vthermal'],
    >>>                     level='l2')
    >>>
    >>> # Load position data
    >>> pyspedas.projects.themis.state(probe='a', trange=trange)
    >>> pos_data = get_data('tha_pos_gsm')
    >>> pos_gsm_time = pos_data.times
    >>> pos_gsm = pos_data.y
    >>>
    >>> # Retrieve ion density
    >>> dens_i_time, dens_i = get_data('tha_peir_density')
    >>>
    >>> # Retrieve spacecraft potential
    >>> sc_pot_time, sc_pot = get_data('tha_peer_sc_pot')
    >>>
    >>> # Retrieve electron velocity
    >>> vth_time, vth = get_data('tha_peer_vthermal')
    >>>
    >>> # Calculate the plasma density from spacecraft potential, ESA datatimes are all the same
    >>> Npot = scpot2dens_nishimura(sc_pot, sc_pot_time, vth, vth_time, dens_i, dens_i_time, pos_gsm, pos_gsm_time, probe='a', no_interp=True)

    References
    ----------
    Based on IDL function THM_SCPOT2DENS_OPT_N.PRO by Toshi Nishimura@UCLA/NAGOYA, 05/02/2009 (toshi at atmos.ucla.edu)
    SPEDAS/projects/themis/common/thm_scpot2dens_opt_n.pro
    """

    # Probe validation and selection
    valid_probes = ['a', 'b', 'c', 'd', 'e']

    if str(probe).lower() not in valid_probes:
        pyspedas.logger.error("Invalid probe entered. Valid probes are [a, b, c, d, e]")
        return None

    # Check if Ni exists and have >= 2 elements if no_interp is False
    if no_interp == False:
        if len(dens_i) < 2 or dens_i[0] == -1:
            pyspedas.logger.error("dens_i must exist and have two or more elements.")
            return None
    else: #Can have a single element if not interpolating
        if dens_i[0] == -1:
            pyspedas.logger.error("dens_i must exist.")
            return None            

# Get position from earth center in KM
    if len(pos_gsm) < 2:
        pyspedas.logger.error("GSM position must exist.")
        return None
    R = np.linalg.norm(pos_gsm, axis = 1)/6371.2
    R_int = interpolate_to_scptime(pos_gsm_time, scptime, R)
    
    if no_interp == False:
        # Interpolate vth and dens_i to scpot time base
        vth_int = interpolate_to_scptime(vth_time, scptime, vth)
        dens_i_int = interpolate_to_scptime(dens_i_time, scptime, dens_i)
    else:
        # Or not
        vth_int = vth
        dens_i_int = dens_i

    #Get vthdata2
    vthdata2 = vth_int*(1-(-1/(1+np.exp(-2*(R_int-4)))+1))+10**4*(-1/(1+np.exp(-2*(R_int-4)))+1)

    #scpn is the negative potential
    scpn = -scpot

    #Get density
    if probe == 'a':
        Ne_scpot = (
            10**(scpn/22.0) * 20000. +
            10**(scpn/ 5.0) * 160000. +
            10**(scpn/ 2.0) * 15000000.0 +
            10**(scpn/ 0.4) * 1.5e12
        ) / vthdata2

    elif probe == 'b':
        Ne_scpot = (
            10**(scpn/25.5) * 16500. +
            10**(scpn/ 2.0) * 3000000.0 +
            10**(scpn/ 0.4) * 1.5e12
        ) / vthdata2

    elif probe == 'c':
        Ne_scpot = (
            10**(scpn/26.0) * 22000. +
            10**(scpn/ 2.0) * 3000000.0 +
            10**(scpn/ 0.25) * 1e14
        ) / vthdata2

    elif probe == 'd':
        Ne_scpot = (
            10**(scpn/25.5) * 20000. +
            10**(scpn/  5.0) * 30000. +
            10**(scpn/  2.0) * 10000000.0 +
            10**(scpn/  0.2) * 5e15
        ) / vthdata2

    elif probe == 'e':
        Ne_scpot = (
            10**(scpn/25.5) * 20000. +
            10**(scpn/  5.0) * 30000. +
            10**(scpn/  2.0) * 10000000.0 +
            10**(scpn/  0.2) * 3e15
        ) / vthdata2

    #That's all, check the ion density
    Ne_scpot[Ne_scpot > ion_density_threshold] = np.nan
    
    return Ne_scpot



import os
import time
import numpy as np
import logging
import requests
import unittest
import subprocess

from pyspedas.tplot_tools import data_exists, del_data, tplot, get_data
from pyspedas.projects.mms import mms_config, mms_load_state,            \
                                  mms_load_tetrahedron_qf, mms_load_mec, \
                                  mms_load_fgm, mms_load_scm,            \
                                  mms_load_hpca, mms_load_edp,           \
                                  mms_load_edi, mms_load_aspoc,          \
                                  mms_load_dsp
                                  # mms_load_fpi, mms_load_feeps
import pyspedas
from pyspedas import tdpwrspc
from pyspedas.projects.mms.hpca_tools.mms_hpca_calc_anodes import mms_hpca_calc_anodes
from pyspedas.projects.mms.hpca_tools.mms_hpca_spin_sum import mms_hpca_spin_sum
from pyspedas.projects.mms.hpca_tools.mms_get_hpca_info import mms_get_hpca_info
#==========================================================

# moto server mock details
localhost = "http://localhost:3000"
bucket_name = "test-bucket"

# Set up mock AWS environment variables (fake credentials)
os.environ["AWS_ACCESS_KEY_ID"] = "test"
os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

# Set environment to use the local Moto S3 server
# S3 ENDPOINT for fsspec
# ENDPOINT URL for cdflib/boto3
os.environ["AWS_S3_ENDPOINT"] = localhost
os.environ["AWS_ENDPOINT_URL"] = localhost

class LoadTestCases(unittest.TestCase):
    """
    Cloud Awareness Unit Tests

    Depends upon moto[server] package. Install via:
        pip install moto[server]

    These tests essentially create a local mock-AWS server as a background
    process at port 3000.
    Note: The environment variables are used as mock credentials in order
          to avoid having to pass the endpoint url to fsspec calls.
    """

    @classmethod
    def setUpClass(cls):
        # Start the moto server for S3 in the background
        # https://github.com/getmoto/moto/issues/4418
        cls.moto_server = subprocess.Popen(
                ["moto_server", "-p3000"],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        # Allow the server to start properly
        time.sleep(2)

        # Create a bucket using direct HTTP requests
        response = requests.put(f"http://localhost:3000/{bucket_name}")
        assert response.status_code == 200, "Bucket creation failed"

    @classmethod
    def tearDownClass(cls):
        # Terminate the moto server after tests
        cls.moto_server.terminate()
        cls.moto_server.communicate()
    
    def clean_data(self):
        # reset moto server to original state
        response = requests.post("http://localhost:3000/moto-api/reset")
        assert response.status_code == 200, "Moto Server reset failed"
        
        # create bucket again
        response = requests.put(f"http://localhost:3000/{bucket_name}")
        assert response.status_code == 200, "Bucket creation failed"

    #==========================================================================
    # Cloud Awareness Note: Cluster implementation does not stream data due to
    #                       lack of functionality already present in PySPEDAS.

    # Adapted unit tests for AWS-specific URI testing.
    def test_state_load_eph_no_update(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_state(datatypes=['pos', 'vel']) # ensure the files are stored locally
        del_data('*') # remove the current tplot vars
        data = mms_load_state(datatypes=['pos', 'vel'], no_update=True) # ensure the files are stored locally
        self.assertTrue(data_exists('mms1_defeph_pos'))
        self.assertTrue(data_exists('mms1_defeph_vel'))
        self.assertTrue('mms1_defeph_pos' in data)
        self.assertTrue('mms1_defeph_vel' in data)
    
    def test_state_load_eph_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_state(datatypes=['pos', 'vel'])
        self.assertTrue(data_exists('mms1_defeph_pos'))
        self.assertTrue(data_exists('mms1_defeph_vel'))
        tplot(['mms1_defeph_pos', 'mms1_defeph_vel'], display=False)
        self.assertTrue('mms1_defeph_pos' in data)
        self.assertTrue('mms1_defeph_vel' in data)
    
    def test_state_load_eph_multiprobe_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_state(datatypes=['pos', 'vel'],probe=['1','2','3','4'])
        self.assertTrue(data_exists('mms1_defeph_pos'))
        self.assertTrue(data_exists('mms1_defeph_vel'))
        tplot(['mms1_defeph_pos', 'mms1_defeph_vel'], display=False)
        self.assertTrue('mms1_defeph_pos' in data)
        self.assertTrue('mms1_defeph_vel' in data)
        self.assertTrue('mms2_defeph_pos' in data)
        self.assertTrue('mms2_defeph_vel' in data)
        self.assertTrue('mms3_defeph_pos' in data)
        self.assertTrue('mms3_defeph_vel' in data)
        self.assertTrue('mms4_defeph_pos' in data)
        self.assertTrue('mms4_defeph_vel' in data)
    
    def test_state_load_tqf(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_tetrahedron_qf()
        self.assertTrue(data_exists('mms_tetrahedron_qf'))
        tplot('mms_tetrahedron_qf',display=False)
        self.assertTrue('mms_tetrahedron_qf' in data)

    def test_state_load_tqf_no_update(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_tetrahedron_qf()  # Ensure that some data is downloaded
        del_data('mms_tetrahedron_qf')    # Delete the tplot variable
        data = mms_load_tetrahedron_qf(no_update=True)  # Ensure that it can be loaded from previously downloaded files
        self.assertTrue(data_exists('mms_tetrahedron_qf'))
        self.assertTrue('mms_tetrahedron_qf' in data)
        tplot('mms_tetrahedron_qf',display=False)
    
    def test_state_load_att_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_state(trange=['2015-10-16', '2015-10-16/06:00'], datatypes=['spinras', 'spindec'])
        self.assertTrue(data_exists('mms1_defatt_spinras'))
        self.assertTrue(data_exists('mms1_defatt_spindec'))
        self.assertTrue('mms1_defatt_spinras' in data)
        self.assertTrue('mms1_defatt_spindec' in data)
        tplot(['mms1_defatt_spinras', 'mms1_defatt_spindec'], display=False)
    
    def test_mec_load_default_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_mec_r_sm'))
        tplot(['mms1_mec_r_sm'], display=False)

    def test_mec_load_spdf_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_mec_r_sm'))

    def test_mec_load_suffix(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_mec(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_mec_r_sm_test'))
    
    def test_fgm_regression_multi_imports_spdf(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        t1, d1 = get_data('mms1_fgm_b_gse_brst_l2')
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        t2, d2 = get_data('mms1_fgm_b_gse_brst_l2')
        self.assertTrue(t1.shape == t2.shape)
        self.assertTrue(d1.shape == d2.shape)

    def test_fgm_load_default_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'],available=True)
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))
        self.assertTrue(data_exists('Epoch'))
        self.assertTrue(data_exists('Epoch_state'))
        tplot(['mms1_fgm_b_gse_srvy_l2'], display=False)

    def test_fgm_load_default_data_exclude(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        # Capture all log messages of level INFO or above
        with self.assertLogs(level=logging.INFO) as captured:
            # assertLogs fails if there are no log messages, so we make sure there's at least one
            logging.info("Dummy log message")
            data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'],exclude_format='*rdeltahalf*',available=True)
            data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'],exclude_format='*rdeltahalf*')
            self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))
            self.assertTrue(data_exists('Epoch'))
            self.assertTrue(data_exists('Epoch_state'))
            tplot(['mms1_fgm_b_gse_srvy_l2'], display=False)
        # Assert that none of the log messages contain the string "rdeltahalf"
        logging.info("Captured log messages:")
        for rec in captured.records:
            logging.info(rec.msg)
            self.assertTrue("rdeltahalf" not in rec.msg)

    def test_fgm_load_spdf_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], spdf=True)
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))

    def test_fgm_load_suffix(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], suffix='_test')
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2_test'))

    def test_fgm_load_multiple_sc(self):
        data = mms_load_fgm(probe=[1, 2, 3, 4], data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms2_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms3_fgm_b_gse_brst_l2'))
        self.assertTrue(data_exists('mms4_fgm_b_gse_brst_l2'))

    def test_fgm_load_brst_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))
        tplot(['mms1_fgm_b_gse_brst_l2'], display=False)

    def test_fgm_load_data_no_update(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00']) # make sure the files exist locally
        del_data('*') 
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'], no_update=True) # load the file from the local cache
        self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))
    
    def test_scm_brst_dpwrspc_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_scm(probe=4, data_rate='brst', datatype='scb', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        tdpwrspc('mms4_scm_acb_gse_scb_brst_l2', notmvariance=True)
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_x_dpwrspc'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_y_dpwrspc'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_z_dpwrspc'))
        tplot(['mms4_scm_acb_gse_scb_brst_l2',
               'mms4_scm_acb_gse_scb_brst_l2_x_dpwrspc',
               'mms4_scm_acb_gse_scb_brst_l2_y_dpwrspc',
               'mms4_scm_acb_gse_scb_brst_l2_z_dpwrspc'], display=False)

    def test_scm_load_default_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2'))

    def test_scm_load_schb(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = pyspedas.mms.scm(probe=4, data_rate='brst', datatype='schb', trange=['2015-10-01/10:48:16', '2015-10-01/10:49:16'])
        self.assertTrue(data_exists('mms4_scm_acb_gse_schb_brst_l2'))

    def test_scm_load_suffix(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2_test'))

    def test_scm_load_multiple_sc(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_scm(probe=['1', '2', '3', '4'], trange=['2017-12-15', '2017-12-16'])
        # self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2'))
        # self.assertTrue(data_exists('mms2_scm_acb_gse_scsrvy_srvy_l2'))
        self.assertTrue(data_exists('mms3_scm_acb_gse_scsrvy_srvy_l2'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scsrvy_srvy_l2'))

    def test_scm_load_brst_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_scm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], datatype='scb')
        self.assertTrue(data_exists('mms1_scm_acb_gse_scb_brst_l2'))

    def test_scm_available(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        files = mms_load_scm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'], datatype='scb', available=True)
        self.assertTrue(len(files) == 2)
    
    def test_hpca_load_default_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_hpca_hplus_number_density'))
        tplot(['mms1_hpca_hplus_number_density'], display=False)

    def test_hpca_load_spdf_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_hpca_hplus_number_density'))
        tplot(['mms1_hpca_hplus_number_density'], display=False)

    def test_hpca_load_ion_omni_suffix(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        del_data('*')
        data = mms_load_hpca(probe=2, trange=['2016-08-09/09:10', '2016-08-09/10:10:00'], datatype='ion', data_rate='brst', suffix='_brst')
        mms_hpca_calc_anodes(fov=[0, 360], probe=2, suffix='_brst')
        mms_hpca_spin_sum(probe=2, suffix='_brst', avg=True)
        self.assertTrue(data_exists('mms2_hpca_hplus_flux_brst_elev_0-360_spin'))
        tplot(['mms2_hpca_hplus_flux_brst_elev_0-360_spin'], display=False)

    def test_hpca_load_ion_omni(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        del_data('*')
        data = mms_load_hpca(trange=['2016-10-16', '2016-10-16/6:00'], datatype='ion')
        mms_hpca_calc_anodes(fov=[0, 360], probe='1')
        mms_hpca_spin_sum()
        self.assertTrue(data_exists('mms1_hpca_hplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_heplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_heplusplus_flux_elev_0-360_spin'))
        self.assertTrue(data_exists('mms1_hpca_oplus_flux_elev_0-360_spin'))
        tplot(['mms1_hpca_hplus_flux_elev_0-360_spin',
               'mms1_hpca_heplus_flux_elev_0-360_spin',
               'mms1_hpca_heplusplus_flux_elev_0-360_spin',
               'mms1_hpca_oplus_flux_elev_0-360_spin'], display=False)

    def test_hpca_center_fast_moments_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_hpca(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
        centered = mms_load_hpca(trange=['2015-10-16/14:00', '2015-10-16/15:00'], center_measurement=True, suffix='_centered')
        
        t, d = get_data('mms1_hpca_hplus_ion_bulk_velocity')
        c, d = get_data('mms1_hpca_hplus_ion_bulk_velocity_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 5.0)

    def test_hpca_center_brst_moments_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_hpca(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst')
        centered = mms_load_hpca(trange=['2015-10-16/13:06', '2015-10-16/13:07'], data_rate='brst', center_measurement=True, suffix='_centered')
        
        t, d = get_data('mms1_hpca_hplus_ion_bulk_velocity')
        c, d = get_data('mms1_hpca_hplus_ion_bulk_velocity_centered')
        self.assertTrue(np.round(c[0]-t[0], decimals=3) == 5.0)

    def test_hpca_info(self):
        info = mms_get_hpca_info()
        self.assertTrue(list(info.keys()) == ['elevation', 't_spin', 't_sweep', 'azimuth_energy_offset'])
    
    def test_edp_load_default_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))
        tplot(['mms1_edp_dce_gse_fast_l2'], display=False)

    def test_edp_load_hfesp_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], datatype='hfesp', data_rate='srvy')
        self.assertTrue(data_exists('mms1_edp_hfesp_srvy_l2'))
        tplot(['mms1_edp_hfesp_srvy_l2'], display=False)

    def test_edp_load_spdf_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))
        tplot(['mms1_edp_dce_gse_fast_l2'], display=False)

    def test_edp_load_suffix(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))

    def test_edp_load_brst_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_edp(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_brst_l2'))
        tplot(['mms1_edp_dce_gse_brst_l2'], display=False)
    
    def test_edi_load_default_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], available=True)
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'])
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2'))
        tplot(['mms1_edi_e_gse_srvy_l2'], display=False)

    def test_edi_load_spdf_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], spdf=True)
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2'))
        tplot(['mms1_edi_e_gse_srvy_l2'], display=False)

    def test_edi_load_suffix(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2_test'))
        tplot(['mms1_edi_e_gse_srvy_l2_test'], display=False)
    
    def test_aspoc_load_default_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], available=True)
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2'))
        tplot(['mms1_aspoc_ionc_l2'], display=False)

    def test_aspoc_load_spdf_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], spdf=True)
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2'))

    def test_aspoc_load_suffix(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'], suffix='_test')
        self.assertTrue(data_exists('mms1_aspoc_ionc_l2_test'))
    
    def test_dsp_load_epsd_bpsd_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype=['epsd', 'bpsd'], level='l2', data_rate='fast')
        self.assertTrue(data_exists('mms1_dsp_epsd_omni'))
        self.assertTrue(data_exists('mms1_dsp_bpsd_omni'))
        tplot(['mms1_dsp_epsd_omni', 'mms1_dsp_bpsd_omni'], display=False)

    def test_dsp_load_bpsd_data(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_dsp(trange=['2015-10-16','2015-10-17'], datatype='bpsd', level='l2', data_rate='fast', available=True)
        data = mms_load_dsp(trange=['2015-10-16','2015-10-17'], datatype='bpsd', level='l2', data_rate='fast')
        self.assertTrue(data_exists('mms1_dsp_bpsd_omni_fast_l2'))
        tplot(['mms1_dsp_bpsd_omni_fast_l2'], display=False)

    def test_dsp_load_epsd_spdf(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype='epsd', level='l2', data_rate='fast', spdf=True)
        self.assertTrue(data_exists('mms1_dsp_epsd_omni'))
        tplot(['mms1_dsp_epsd_omni'], display=False)

    def test_dsp_load_epsd_suffix(self):
        self.clean_data()
        mms_config.CONFIG['local_data_dir'] = f"s3://{bucket_name}"

        data = mms_load_dsp(trange=['2015-08-01','2015-08-02'], datatype='epsd', level='l2', data_rate='fast', suffix='_test')
        self.assertTrue(data_exists('mms1_dsp_epsd_omni_test'))
        tplot(['mms1_dsp_epsd_omni_test'], display=False)

if __name__ == '__main__':
    unittest.main()



import unittest
import logging
import pyspedas
from pyspedas.projects.erg.satellite.erg.particle.erg_lepe_get_dist import erg_lepe_get_dist
from pyspedas.particles.spd_part_products.spd_pgs_make_theta_spec import spd_pgs_make_theta_spec
from pyspedas.particles.spd_part_products.spd_pgs_make_phi_spec import spd_pgs_make_phi_spec
from pyspedas.particles.spd_part_products.spd_pgs_make_e_spec import spd_pgs_make_e_spec

class MyTestCase(unittest.TestCase):
    def test_theta_spec(self):
        trange=['2017-04-05 21:45:00', '2017-04-05 22:45:00']
        pyspedas.projects.erg.lepe(trange=trange,datatype='3dflux')
        times=erg_lepe_get_dist('erg_lepe_l2_3dflux_FEDU', trange=['2017-04-05 21:45:00', '2017-04-05 22:45:00'], time_only=True)
        # spd_pgs_make_theta_spec does bin weighting, which is horribly slow in practice.  So we'll just do a single distribution.
        single_time=times[0]
        dist=erg_lepe_get_dist('erg_lepe_l2_3dflux_FEDU', single_time=single_time)
        self.assertTrue(dist is not None)
        theta_vals, spectra = spd_pgs_make_theta_spec(dist)
        self.assertTrue(theta_vals is not None)
        self.assertTrue(spectra is not None)

    def test_phi_spec(self):
        trange=['2017-04-05 21:45:00', '2017-04-05 22:45:00']
        pyspedas.projects.erg.lepe(trange=trange,datatype='3dflux')
        times=erg_lepe_get_dist('erg_lepe_l2_3dflux_FEDU', trange=['2017-04-05 21:45:00', '2017-04-05 22:45:00'], time_only=True)
        # spd_pgs_make_phi_spec does bin weighting, which is horribly slow in practice.  So we'll just do a single distribution.
        single_time=times[0]
        dist=erg_lepe_get_dist('erg_lepe_l2_3dflux_FEDU', single_time=single_time)
        self.assertTrue(dist is not None)
        phi_vals, spectra = spd_pgs_make_phi_spec(dist)
        self.assertTrue(phi_vals is not None)
        self.assertTrue(spectra is not None)

    def test_e_spec(self):
        trange=['2017-04-05 21:45:00', '2017-04-05 22:45:00']
        pyspedas.projects.erg.lepe(trange=trange,datatype='3dflux')
        times=erg_lepe_get_dist('erg_lepe_l2_3dflux_FEDU', trange=['2017-04-05 21:45:00', '2017-04-05 22:45:00'], time_only=True)
        dist=erg_lepe_get_dist('erg_lepe_l2_3dflux_FEDU', trange=trange)
        self.assertTrue(dist is not None)
        e_vals, spectra = spd_pgs_make_e_spec(dist)
        self.assertTrue(e_vals is not None)
        self.assertTrue(spectra is not None)




if __name__ == '__main__':
    unittest.main()



from pyspedas.projects.mms.mec_ascii.mms_get_tetrahedron_qf import mms_get_tetrahedron_qf



def mms_load_tetrahedron_qf(trange=['2015-10-16', '2015-10-17'], no_update=False, suffix=''):
    """
    Load the MMS tetrahedron quality factor data from the ASCII files
    into tplot variables
    
    Parameters
    ----------
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2015-10-16', '2015-10-17']

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.
            Default: None

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten
            Default: False

    Returns
    --------
        list of str
            List of tplot variables created.

    Example
    -------

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> qf_vars = pyspedas.projects.mms.mms_load_tetrahedron_qf(trange=['2015-10-16', '2015-10-17'])
    >>> tplot('mms_tetrahedron_qf')

    """
    return mms_get_tetrahedron_qf(trange=trange, no_download=no_update, suffix=suffix)



import bisect
import logging
import numpy as np
from pyspedas.tplot_tools import get_data, store_data, options
from pyspedas.tplot_tools import degap


def spec_pa_sort(
    spec2plot,
    pas2plot,
):
    """
    Sort pitch angle in ascending order. 
    If pitch angle are not monotonic, can't plot it with tplot.

    Parameters
    ----------
        spec2plot: ndarray
            spectogram variable, can be 2d or 3d array. 
        pas2plot: ndarray
            pitch angle variable, should be the same size as spec_var, or the size of a slice of spec_var

    Return
    ---------
        spec2plot: in sorted order
        pas2plot: in sorted order
    """
    if len(spec2plot.shape) == 2:
        spec_sectNum, spec_paNum = spec2plot.shape
        nEngChannel = 1

    elif len(spec2plot.shape) == 3:
        spec_sectNum, spec_paNum, nEngChannel = spec2plot.shape
    else:
        logging.error("ELF EPD L2 PA SPECTOGRAM: spec2plot is not 2d or 3d!")
        return

    if len(pas2plot.shape) == 2:
        pa_sectNum, pa_paNum = pas2plot.shape
    else:
        logging.error("ELF EPD L2 PA SPECTOGRAM: pas2plot is not 2d!")
        return    

    if (pa_sectNum != spec_sectNum) or (pa_paNum != spec_paNum):
        logging.error("ELF EPD L2 PA SPECTOGRAM: pas2plot and spec2plot not same size!")
        return
    
    if len(spec2plot.shape) == 3:
        for i in range(spec_sectNum):
            line = pas2plot[i, :]
            # If the line is in descending order
            if np.array_equal(line, np.sort(line)[::-1]):
                # Flip the line
                pas2plot[i, :] = line[::-1]
                for j in range(nEngChannel):
                    spec2plot[i, :, j] = spec2plot[i, :, j][::-1]
    elif len(spec2plot.shape) == 2:
        for i in range(spec_sectNum):
            line = pas2plot[i, :]
            # If the line is in descending order
            if np.array_equal(line, np.sort(line)[::-1]):
                # Flip the line
                pas2plot[i, :] = line[::-1]
                spec2plot[i, :] = spec2plot[i, :][::-1]

    return spec2plot, pas2plot


def epd_l2_Espectra_option(
    flux_var,
):
    """
    Add options for omni/para/anti/perp flux spectra

    Parameters
    ----------
        flux_var: str
            Tplot variable name of 2d omni/para/anti/perp flux spectra

    """
    unit_ = '#/(s-cm$^2$\n-str-MeV)' if "nflux" in flux_var else 'keV/(s-cm$^2$\n-str-MeV)'
    zrange = [10, 2e7] if "nflux" in flux_var else [1.e1, 2.e7]

    options(flux_var, 'spec', True)
    options(flux_var, 'yrange', [55., 6800])  # energy range
    options(flux_var, 'zrange', zrange)
    options(flux_var, 'ylog', True)
    options(flux_var, 'zlog', True)
    flux_var_ytitle = flux_var[0:10] + "\n" + flux_var[11:]
    options(flux_var, 'ytitle', flux_var_ytitle)
    options(flux_var, 'ysubtitle', '[keV]')
    options(flux_var, 'ztitle', unit_)

    
def epd_l2_Espectra(
    flux_tvar, 
    LC_tvar,
    LCfatol=None,
    LCfptol=None,
    nodegap=True,
    ):

    """
    Produce omni/para/perp/anti flux spectra for ELF EPD L2 data

    Parameters
    ----------
        flux_tvar: str
            Tplot variable name of a 3d energy time spectra 
        LC_tvar: str
            Tplot variable name of loss cone
        LCfatol: float, optional
            Tolerance angle for para and anti flux. A positive value makes the loss 
            cone/antiloss cone smaller by this amount. 
            Default is 22.25 deg.
        LCfptol: float, optional
            Tolerance angle for perp flux. A negative value means a wider angle for 
            perp flux.
            Default is -11 deg.
        Nodegap: bool, optional
            Flag for degap. If set, skip degap. 
            Default is False.

    Return
    ----------
        List of four elements.
            omni_var: str. Tplot variable name of 2d omni spectra
            para_var: str. Tplot variable name of 2d parallel spectra
            perp_var: str. Tplot variable name of 2d perpendicualr spectra
            anti_var: str. Tplot variable name of 2d antiparallel spectra
    """
    
    # load L2 t-PA-E 3D flux 
    data = get_data(flux_tvar)

    # parameters setup
    nspinsavailable, nPAsChannel, nEngChannel= np.shape(data.y)
    if 'hs' in flux_tvar:
        nspinsectors = (nPAsChannel - 2)*2 
    elif 'fs' in flux_tvar:
        nspinsectors = (nPAsChannel - 2)
    else:
        logging.error("ELF EPD L2 ENERGY SPECTROGRAM: invalid input tplot variable!")

    FOVo2 = 11.  # Field of View divided by 2 (deg)
    dphsect = 360./nspinsectors 
    SectWidtho2 = dphsect/2.
    LCfatol = FOVo2 + SectWidtho2  if LCfatol is None else LCfatol # tolerance of pitch angle in field aligned direction, default 22.25 deg
    LCfptol = -FOVo2  if LCfptol is None else LCfptol # tolerance of pitch angel in perpendicular direction, default -11. deg
    # TODO: make sure these default values are correct
    
    # calculate domega in PA
    pas2plot = np.copy(data.v1)
    spec2plot = np.copy(data.y)
    energy = np.copy(data.v2)
    pas2plot_domega = (2. * np.pi / nspinsectors) * np.sin(np.pi * pas2plot / 180.)
    index1, index2 = np.where((pas2plot < 360. / nspinsectors / 2) | (pas2plot > 180. - 360. / nspinsectors / 2))
    pas2plot_domega[index1, index2] = (np.pi/nspinsectors)*np.sin(np.pi/nspinsectors) 

    # # pas2plot_bcast need to have the same nan as spec2plot. 
    # if not, np.nansum(pas2plot_bcast, axis=1) will be wrong
    pas2plot_domega[:, 0] = np.nan
    pas2plot_domega[:, -1] = np.nan
    #pas2plot_repeat = np.repeat(pas2plot[:, :, np.newaxis], nEngChannel, axis=2)
    pas2plot_domega_bcast = np.broadcast_to(pas2plot_domega[:, :, np.newaxis], (nspinsavailable, nPAsChannel, nEngChannel))

    #===========================
    #        OMNI FLUX
    #=========================== 
    # calculate omni flux
    Espectra_omni = np.nansum(spec2plot * pas2plot_domega_bcast, axis=1) / np.nansum(pas2plot_domega_bcast, axis=1)

    # output omni tvar
    omni_var = f"{flux_tvar.replace('Epat_','')}_omni"
    store_data(omni_var, data={'x': data.times, 'y': Espectra_omni, 'v': energy}, attr_dict=get_data(flux_tvar, metadata=True))
    epd_l2_Espectra_option(omni_var)

    #===========================
    #        PARA FLUX
    #=========================== 
    # load loss cone data
    data = get_data(LC_tvar)

    # loss cone
    paraedgedeg = np.array([lc if lc < 90 else 180-lc for lc in data.y])
    paraedgedeg_bcast = np.broadcast_to(paraedgedeg[:, np.newaxis], (nspinsavailable, nPAsChannel))

    # select index 
    iparapas, jparapas = np.where(pas2plot< paraedgedeg_bcast-LCfatol)
    spec2plot_allowable = np.zeros((nspinsavailable, nPAsChannel, nEngChannel))
    spec2plot_allowable[iparapas, jparapas, :] = 1
    Espectra_para = np.nansum(spec2plot * pas2plot_domega_bcast * spec2plot_allowable, axis=1) / np.nansum(pas2plot_domega_bcast * spec2plot_allowable, axis=1)
 
    # output para tvar
    para_var = f"{flux_tvar.replace('Epat_','')}_para"
    store_data(para_var, data={'x': data.times, 'y': Espectra_para, 'v': energy}, attr_dict=get_data(flux_tvar, metadata=True))
    epd_l2_Espectra_option(para_var)

    #===========================
    #        ANTI FLUX
    #=========================== 
    # select index 
    iantipas, jantipas = np.where(pas2plot > 180+LCfatol-paraedgedeg_bcast)
    spec2plot_allowable = np.zeros((nspinsavailable, nPAsChannel, nEngChannel))
    spec2plot_allowable[iantipas, jantipas, :] = 1
    Espectra_anti = np.nansum(spec2plot * pas2plot_domega_bcast * spec2plot_allowable, axis=1) / np.nansum(pas2plot_domega_bcast * spec2plot_allowable, axis=1)

    # output anti tvar
    anti_var = f"{flux_tvar.replace('Epat_','')}_anti"
    store_data(anti_var, data={'x': data.times, 'y': Espectra_anti, 'v': energy}, attr_dict=get_data(flux_tvar, metadata=True))
    epd_l2_Espectra_option(anti_var)

    #===========================
    #        PERP FLUX
    #=========================== 
    # select index 
    iperppas, jperppas = np.where(
        (pas2plot < 180-LCfptol-paraedgedeg_bcast) & (pas2plot > LCfptol+paraedgedeg_bcast))
    spec2plot_allowable = np.zeros((nspinsavailable, nPAsChannel, nEngChannel))
    spec2plot_allowable[iperppas, jperppas, :] = 1
    Espectra_perp = np.nansum(spec2plot * pas2plot_domega_bcast * spec2plot_allowable, axis=1) / np.nansum(pas2plot_domega_bcast * spec2plot_allowable, axis=1)

    # output anti tvar
    perp_var = f"{flux_tvar.replace('Epat_','')}_perp"
    store_data(perp_var, data={'x': data.times, 'y': Espectra_perp, 'v': energy}, attr_dict=get_data(flux_tvar, metadata=True))
    epd_l2_Espectra_option(perp_var)
    
    #===========================
    #        DEGAP
    #=========================== 
    if nodegap is False:
        nspinsinsum = get_data(f"{flux_tvar[0:3]}_pef_nspinsinsum")
        spinper = get_data(f"{flux_tvar[0:3]}_pef_tspin")
        if np.average(nspinsinsum.y) > 1 :
            mydt = np.max(nspinsinsum.y)*np.median(spinper.y)
        else:
            mydt = np.median(spinper.y)

        degap(omni_var, dt=mydt, margin=0.5*mydt/2)
        degap(para_var, dt=mydt, margin=0.5*mydt/2)
        degap(anti_var, dt=mydt, margin=0.5*mydt/2)
        degap(perp_var, dt=mydt, margin=0.5*mydt/2)


    return [omni_var, para_var, anti_var, perp_var]


def epd_l2_PAspectra_option(
    flux_var,
    set_zrange = True,
):
    """
    Add options for pitch angle spectra

    Parameters
    ----------
        flux_var: str
            Tplot variable name of 2d pitch angle spectra

        set_zrange: bool, optonal 
            Default is True. Only works for default four channels.
    """

    unit_ = '#/(s-cm$^2$\n-str-MeV)' if "nflux" in flux_var else 'keV/(s-cm$^2$\n-str-MeV)'
    if set_zrange is True:
        if "nflux" in flux_var:
            zrange_list = {
                0: [2.e3, 2.e7],
                1: [1.e3, 4.e6],
                2: [1.e2, 1.e6],
                3: [1.e1, 2.e4],
            }
        else:
            zrange_list = {
                0: [5.e5, 1.e10],
                1: [5.e5, 5.e9],
                2: [5.e5, 2.5e9],
                3: [1.e5, 1.e8],
            }
        ch_num = int(flux_var.split("ch")[-1])
        zrange = zrange_list.get(ch_num, [1e1, 1e2]) if "nflux" in flux_var else zrange_list.get(ch_num, [1.e4, 1.e7])
        options(flux_var, 'zrange', zrange)

    options(flux_var, 'spec', True)
    #options(flux_var, 'yrange', [-180, 180])  # energy range
    options(flux_var, 'ylog', False)
    options(flux_var, 'zlog', True)
    flux_var_ytitle = flux_var[0:10] + "\n" + flux_var[11:]
    options(flux_var, 'ytitle', flux_var_ytitle)
    options(flux_var, 'ysubtitle', 'PA (deg)')
    options(flux_var, 'ztitle', unit_)


def epd_l2_PAspectra(
    flux_tvar, 
    energybins = None,
    energies = None,
    nodegap=True,
    ):
    """
    This function processes a 3D energy-time spectra from ELF EPD L2 data 
    to produce pitch angle spectra, applying either specified energy bins 
    or a defined energy range. 

    Parameters
    ----------
        flux_tvar: str
            Tplot variable name of a 3d energy time spectra 

        energybins: list of int, optional
            Specified the energy bins used for generating pitch angle spectra. 
            Default is [(0,2),(3,5), (6,8), (9,15)], which bins energy as follows:
                ch0: Energy channels 0-2,
                ch1: Energy channels 3-5, 
                ch2: Energy channels 6-8, 
                ch3: Energy channels 9-15
            If both 'energybins' and 'energies' are set, 'energybins' takes precedence
            energy 

        energies: list of tuple of float, optional
            Specifies the energy range for each bin in the pitch angle spectra.
            Example: energies=[(50.,160.),(160.,345.),(345.,900.),(900.,7000.)]
            If both 'energybins' and 'energies' are set, 'energybins' takes precedence.
            Energy and energybin table:
            channel     energy_range    energy_midbin
            0           50-80           63.2
            1           80-120          97.9
            2           120-160         138.5
            3           160-210         183.3
            4           210-270         238.1
            5           270-345         305.2
            6           345-430         385.1
            7           430-630         520.4
            8           630-900         752.9
            9           900-1300        1081.6
            10          1300-1800       1529.7
            11          1800-2500       2121.3
            12          2500-3350       2893.9
            13          3350-4150       3728.6
            14          4150-5800       4906.1
            15          5800+           6500.0
        Nodegap: bool, optional
            Flag for degap. If set, skip degap. 
            Default is False.

    Return
    ----------
        List of tplot variable names for pitch angle spectra
    """
    # constant energy range
    EMINS = np.array([
        50.000008, 79.999962, 120.00005, 159.99998, 210.00015, 269.99973,
        345.00043, 429.99945, 630.00061, 899.99890, 1300.0013, 1799.9985,
        2500.0022, 3349.9990, 4150.0034, 5800.0000])  # TODO: make sure it's safe to hardcode here
    EMAXS = np.array([
        79.999962, 120.00005, 159.99998, 210.00015, 269.99973, 345.00043,
        429.99945, 630.00061, 899.99890, 1300.0013, 1799.9985, 2500.0022,
        3349.9990, 4150.0034, 5800.0000, 7200.0000])

    # load L2 t-PA-E 3D flux 
    data = get_data(flux_tvar)
    energy_midbin = np.copy(data.v2)
    pas2plot = np.copy(data.v1)
    spec2plot = np.copy(data.y)
    nspinsavailable, nPAsChannel, nEngChannel= np.shape(data.y)
    #nspinsectors = (nPAsChannel - 2)*2 

    spec2plot, pas2plot = spec_pa_sort(spec2plot, pas2plot)
    
    # get energy bin for pitch angle spectra
    if energybins is not None:
        MinE_channels, MaxE_channels = zip(*energybins)
        MinE_channels = list(MinE_channels)
        MaxE_channels = list(MaxE_channels)
        if energies is not None:
            logging.warning("both energis and energybins are set, 'energybins' takes precedence!")  
    elif energies is not None:
        MinE_channels = [bisect.bisect_left(energy_midbin, min_energy) for min_energy, max_energy in energies]
        MaxE_channels = [bisect.bisect_right(energy_midbin, max_energy) - 1 for min_energy, max_energy in energies]
    else:
        MinE_channels = [0, 3, 6, 9]
        MaxE_channels = [2, 5, 8, 15]
    logging.info(f"Energy channel {list(zip(MinE_channels, MaxE_channels))} are used for epd l2 pitch angle spectra.") 

    # broadcast Emax and Emin so that they have the same size of spec2plot    
    Emaxs_bcast = np.broadcast_to(EMAXS[np.newaxis, np.newaxis, :], (nspinsavailable, nPAsChannel, nEngChannel))
    Emins_bcast = np.broadcast_to(EMINS[np.newaxis, np.newaxis, :], (nspinsavailable, nPAsChannel, nEngChannel))
    
    # define PAspectra, size is time * pa * energy channel
    numchannels = len(MinE_channels)
    PAspectra = np.full((nspinsavailable, nPAsChannel, numchannels), np.nan)
    PA_tvars = []
    
    # loop over specified energy channels
    for ichannel in range(numchannels):
        min_channel =  MinE_channels[ichannel]
        max_channel =  MaxE_channels[ichannel]
        
        PAspectra_single = np.nansum(spec2plot[:,:,min_channel:max_channel+1] * \
                    (Emaxs_bcast[:,:,min_channel:max_channel+1] - Emins_bcast[:,:,min_channel:max_channel+1]), axis=2) \
                    / np.nansum(Emaxs_bcast[:,:,min_channel:max_channel+1] - Emins_bcast[:,:,min_channel:max_channel+1], axis=2)
        
        # make the first and last pa nan, for the comparison with idl
        PAspectra_single[:, 0] = np.nan
        PAspectra_single[:, -1] = np.nan

        # output tvariable
        PA_var = f"{flux_tvar.replace('Epat_','')}_ch{ichannel}"

        store_data(PA_var, data={'x': data.times, 'y': PAspectra_single, 'v': pas2plot})
        epd_l2_PAspectra_option(PA_var)

        #===========================
        #        DEGAP
        #=========================== 
        if nodegap is False:
            nspinsinsum = get_data(f"{flux_tvar[0:3]}_pef_nspinsinsum")
            spinper = get_data(f"{flux_tvar[0:3]}_pef_tspin")
            if np.average(nspinsinsum.y) > 1 :
                mydt = np.max(nspinsinsum.y)*np.median(spinper.y)
            else:
                mydt = np.median(spinper.y)
            degap(PA_var, dt=mydt, margin=0.5*mydt/2)
        
        PA_tvars.append(PA_var)


    return PA_tvars


import logging
import unittest
from pyspedas.projects.themis import state, ssc, ssc_pre
from pyspedas.tplot_tools import data_exists, get_data, del_data, tplot_restore
from numpy.testing import assert_allclose

class Themis_StateDataTests(unittest.TestCase):
    """ Tests creation of support variables in themis.state() """

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The IDL script that creates data file: general/tools/python_validate/thm_state_validate.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.projects.themis.config import CONFIG

        # Testing time range
        cls.t = ['2008-03-23', '2008-03-28']


        # Download validation file
        remote_server = 'https://github.com/spedas/test_data/raw/refs/heads/main/'
        remote_name = 'load_routine_tests/thm_state_validate.tplot'
        datafile = download(remote_file=remote_name,
                           remote_path=remote_server,
                           local_path=CONFIG['local_data_dir'],
                           no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        tplot_restore(filename)
        #tplot_names()
        cls.tha_pos = get_data('tha_state_pos')
        cls.tha_vel = get_data('tha_state_vel')
        cls.tha_spinras = get_data('tha_state_spinras')
        cls.tha_spindec = get_data('tha_state_spindec')
        cls.tha_spinras_correction = get_data('tha_state_spinras_correction')
        cls.tha_spindec_correction = get_data('tha_state_spindec_correction')
        cls.tha_spinras_corrected = get_data('tha_state_spinras_corrected')
        cls.tha_spindec_corrected = get_data('tha_state_spindec_corrected')

        # Load with pyspedas
        state(probe='a',trange=cls.t,get_support_data=True)

    def setUp(self):
        """ We need to clean tplot variables before each run"""
        # del_data('*')

    def test_state_spinras(self):
        """Validate state variables """
        my_data = get_data('tha_state_spinras')
        assert_allclose(my_data.y,self.tha_spinras.y,rtol=1.0e-06)

    def test_state_spindec(self):
        """Validate state variables """
        my_data = get_data('tha_state_spindec')
        assert_allclose(my_data.y,self.tha_spindec.y,rtol=1.0e-06)

    def test_state_spinras_correction(self):
        """Validate state variables """
        my_data = get_data('tha_state_spinras_correction')
        assert_allclose(my_data.y,self.tha_spinras_correction.y,rtol=1.0e-06)

    def test_state_spindec_correction(self):
        """Validate state variables """
        my_data = get_data('tha_state_spindec_correction')
        assert_allclose(my_data.y,self.tha_spindec_correction.y,rtol=1.0e-06)

    def test_state_spinras_corrected(self):
        """Validate state variables """
        my_data = get_data('tha_state_spinras_corrected')
        assert_allclose(my_data.y,self.tha_spinras_corrected.y,rtol=1.0e-06)

    def test_state_spindec_corrected(self):
        """Validate state variables """
        my_data = get_data('tha_state_spindec_corrected')
        assert_allclose(my_data.y,self.tha_spindec_corrected.y,rtol=1.0e-06)

    def test_state_reload_no_v03(self):
        # Test overwriting of spin axis correction variables if data is loaded with V03 corrections, then other
        # data loaded that doesn't have the corrections (prevents dangling correction variables)
        ts1 = ['2007-03-23','2007-03-24']
        ts2 = ['2023-01-01','2023-01-02']
        state(trange=ts1,probe='a',get_support_data=True) # V03 corrections exist
        self.assertTrue(data_exists('tha_spinras_correction'))
        self.assertTrue(data_exists('tha_spindec_correction'))
        state(trange=ts2,probe='a',get_support_data=True) # V03 corrections do not exist
        self.assertFalse(data_exists('tha_spinras_correction'))
        self.assertFalse(data_exists('tha_spindec_correction'))


    def test_state_exclude_format(self):
        # Test that the exclude_format option to state() works
        state(trange=['2007-03-23','2007-03-24'], probe='b',varformat='*pos*',exclude_format='*sse*')
        self.assertTrue(data_exists('thb_pos_gse'))
        self.assertFalse(data_exists('thb_pos_sse'))

    def test_state_suffix_spinmodel(self):
        from pyspedas.projects.themis import get_spinmodel
        # Test that the exclude_format option to state() works
        state(trange=['2007-03-23','2007-03-24'], probe='a',suffix='_suffix', get_support_data=True)
        model = get_spinmodel('a', 2)
        self.assertTrue(model is not None)

    def test_state_spdf(self):
        from pyspedas.projects.themis.config import CONFIG
        del_data('the*')
        save = CONFIG['remote_data_dir']
        CONFIG['remote_data_dir'] = 'https://spdf.gsfc.nasa.gov/pub/data/themis/'
        state(trange=['2007-04-01','2007-04-02'], probe='e')
        CONFIG['remote_data_dir'] = save
        self.assertTrue(data_exists('tha_pos'))

    def test_state_ucla(self):
        from pyspedas.projects.themis.config import CONFIG
        del_data('the*')
        save = CONFIG['remote_data_dir']
        CONFIG['remote_data_dir'] = 'https://themis-data.igpp.ucla.edu/'
        state(trange=['2009-04-01','2009-04-02'], probe='e')
        CONFIG['remote_data_dir'] = save
        self.assertTrue(data_exists('tha_pos'))

    def test_ssc(self):
        vars = ssc()
        self.assertTrue(len(vars) > 0)

    def test_ssc_pre(self):
        vars = ssc_pre()
        self.assertTrue(len(vars) > 0)

if __name__ == '__main__':
    unittest.main()



def rate_connection_quality(transfer_time: float, transfer_size:float, transfer_rate:float) -> str:
    """
    Rate the connection quality in terms of the transfer size, transfer time, and transfer rate

    Parameters
    ----------
    transfer_time: float
        Elapsed time for the download, seconds
    transfer_size: float
        Amount of data transferred, megabytes
    transfer_rate: float
         Speed of the transfer, megabytes/sec

    Returns
    -------
    string
    "transfer_slow" if the connection seems slower than expected, "transfer_normal" otherwise

    """
    # We may want to add more quality tiers, but this will serve as a proof of concept
    # Don't bother rating small or very fast transfers
    if transfer_time < 2.0:
        return 'transfer_normal'
    elif transfer_rate < 1.0:
        return 'transfer_slow'
    else:
        return 'transfer_normal'


import unittest
from io import StringIO
import sys
import pyspedas
import pyspedas
from pyspedas.utilities.libs import libs


class LibsTestCase(unittest.TestCase):

    def setUp(self):

        # We do not need this anymore:
        # # Call function once to import all the submodules to silence the output
        # try:
        #     libs('')
        # except Exception as e:
        #     self.fail("Unexpected exception %s" % e)

        # Redirect stdout to capture print statements
        self.held_stdout = sys.stdout
        sys.stdout = StringIO()


    def tearDown(self):
        # Reset stdout
        sys.stdout = self.held_stdout

    def test_known_function_fgm(self):
        libs('fgm')
        output = sys.stdout.getvalue()
        self.assertIn('fgm', output)
        self.assertIn('Function:', output)

    def test_non_existent_function(self):
        libs('non_existent_function')
        output = sys.stdout.getvalue()
        self.assertEqual(output, '')  # Assuming no output for non-existent functions

    def test_partial_function_name(self):
        libs('mms_')
        output = sys.stdout.getvalue()
        self.assertIn('mms_', output)

    def test_wildcard_function_name(self):
        libs('wa*pol')
        output = sys.stdout.getvalue()
        self.assertIn('twavpol', output)

    def test_wildcard_partial_wrapper(self):
        libs('ep*ad')
        output = sys.stdout.getvalue()
        self.assertIn('epead', output)

    def test_substring_partial_wrapper(self):
        libs('epead')
        output = sys.stdout.getvalue()
        self.assertIn('epead', output)

    def test_known_function_subpackage(self):
        libs('fgm', package=pyspedas.projects.themis)
        output = sys.stdout.getvalue()
        self.assertIn('fgm', output)
        self.assertNotIn('mms', output)

    def test_known_function_pytplot(self):
        libs('get_data')
        output = sys.stdout.getvalue()
        self.assertIn('get_data', output)
        self.assertIn('Function:', output)

    def test_known_function_version_pyspeds_subpackage(self):
        libs('version', package=pyspedas)
        output = sys.stdout.getvalue()
        self.assertIn('version', output)
        self.assertNotIn('Function: pytplot', output)

    def test_known_function_version_pyspeds_themis_subpackage(self):
        libs('fgm', package=pyspedas.projects.themis)
        output = sys.stdout.getvalue()
        self.assertIn('fgm', output)

    def test_known_function_pytplot_get_data(self):
        libs('get_data')
        output = sys.stdout.getvalue()
        self.assertIn('get_data', output)
        self.assertIn('Function:', output)

    def test_known_function_data_pytplot_subpackage_only(self):
        libs('data', package=pyspedas)
        output = sys.stdout.getvalue()
        self.assertIn('get_data', output)
        self.assertIn('Function: pyspedas', output)

    def test_qtplotter_error_exception(self):
        # This test is probably not the best way to handle this exception
        libs('qtplotter', package=pyspedas)  # This can be changed to anything. qtplotter error is during pytplot import search
        output = sys.stdout.getvalue()
        self.assertNotIn('Error importing module pytplot.QtPlotter', output)

if __name__ == '__main__':
    unittest.main()
